#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov  5 16:37:22 2019

@author: routhier
"""

import numpy as np
import pyBigWig
import warnings


from kipoiseq.transforms.functional import resize_interval


from .utils import rolling_window
from .normalization import Normalizer, BiNormalizer


class bbi_extractor(object):
    """
    Reads the data into a bigWig file. Returns the coverage on an interval
    with the desired normalization and sampling strategy.

    args:
        bbi_files:
            List of bbi_files from which the data will be taken.
        window:
            The desired length of the output window.
        nb_annotation_type:
            The number of different annotation in the bbi_files. The same
            number of files must be passed for every annotation, the list must
            be organised as [file1_ann1, file1_ann2, file2_ann1, file2_ann2],
            with file1, file2 designing two differents kind of files (different
            lab, different cellular type ...). 
            If None the output shape will be (window, nb_of_files).
        sampling_mode:
            {None, 'mean', 'downsampling'}, the sampling strategy to apply
            at the coverage.
        normalization_mode:
            argument from the Normalizer class
        *args, **kwargs:
            other arguments from Normalizer class
    """
    def __init__(self,
                 bbi_files,
                 window,
                 nb_annotation_type=None,
                 sampling_mode=None,
                 normalization_mode=None,
                 *args,
                 **kwargs):
        if not isinstance(bbi_files, list):
            self.bbi_files = [bbi_files]
        else:
            self.bbi_files = bbi_files

        self.window = window
        self.nb_annotation_type = nb_annotation_type
        self.sampling_mode = sampling_mode
        self.normalization_mode = normalization_mode

        #TODO create a bam, bedGraph, wig to bigWig converter
        self.norm_dico = dict()
        if isinstance(self.normalization_mode, list):
            for bbi_file in self.bbi_files:
                self.norm_dico[bbi_file] = BiNormalizer(normalization_mode,
                                                        bbi_file,
                                                        *args,
                                                        **kwargs)
        else:
            for bbi_file in self.bbi_files:
                self.norm_dico[bbi_file] = Normalizer(normalization_mode,
                                                      bbi_file,
                                                      *args,
                                                      **kwargs)

        if self.nb_annotation_type:
            assert len(self.bbi_files) % self.nb_annotation_type == 0,\
            """Every annotation must be described by the same number of file"""
            
            warnings.warn("""the list must be organised as [file1_ann1, file1_ann2,
            file2_ann1, file2_ann2], with file1, file2 designing two differents kind
            of files (different lab, different cellular type ...)""")

    def extract(self, interval):
        """
        Extract the coverage corresponding to the interval from the bbi_files.
        
        returns:
            np.array of shape (window,
                               number of files per annotation,
                               number of annotation)
        """
        if not self.sampling_mode:
            assert self.window <= abs(interval.stop - interval.start),\
            """The target window must be smaller than the input length"""

            interval = resize_interval(interval,
                                       self.window,
                                       anchor='center')
        seq = list()
        for bbi_file in self.bbi_files:
            bw = pyBigWig.open(bbi_file)
            array = bw.values(interval.chrom,
                              interval.start,
                              interval.stop, numpy=True)
            array[np.isnan(array)] = 0
            seq.append(self.norm_dico[bbi_file](array))
        seq = np.array(seq).T

        if self.sampling_mode:
            assert abs(interval.stop - interval.start) % self.window == 0,\
            """Window must divide the input length to use downsampling"""
            sampling_length = abs(interval.stop - interval.start) // self.window

            if self.sampling_mode == 'downsampling':
                    seq = seq[::sampling_length]
            elif self.sampling_mode == 'mean':
                    seq = self._calculate_rolling_mean(seq)
            else:
                raise NameError('sampling_mode must be None, "mean" or "downsampling"')
        
        if self.nb_annotation_type:
            nb_files_per_ann = len(self.bbi_files) // self.nb_annotation_type
            return seq.reshape((self.window,
                                nb_files_per_ann,
                                self.nb_annotation_type))
        
        else:
            return seq

    def _calculate_rolling_mean(self, x):
        sampling_length = len(x) // self.window
        num_classes = x.shape[1]
        x = rolling_window(x,
                           window=(sampling_length, num_classes),
                           asteps=(sampling_length, num_classes))
        x = x.reshape((self.window, sampling_length, num_classes))
        x = np.mean(x, axis=1, dtype='float32')
        return x


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun 18 14:21:30 2019

@author: routhier
"""

import numpy as np
import inspect
from copy import deepcopy

from .sequence import SeqIntervalDl, StringSeqIntervalDl
from .normalization import Weights
from .utils import ArgumentsDict, get_default_args

class Generator(object):
    """
    info:
         doc: >
             Generator for keras model able to yield inputs and targets in batch
             by reading into a fasta file and a annotation file. Inputs are one-
             hot-encoded or string.
     args:
         batch_size:
             number of example per batch pass to the model.
         one-hot-encoding:
             whether or not the inputs is one-hot-encoded (False: string)
             default: True
         output_shape:
             How to modify the shape of the output (because the initial output
             structure is (batch, length, nb_types, nb_annotation) or (batch,
             nb_types, nb_annotation))
         weighting_mode:
             {None, 'balanced', tuple(weights, bins), tuple(weight_pos, weight_neg)}
             the methodology to set the weights.
             For continuous dataset a tuple with the weights to apply and the
             bins can be parsed (len(weights) == len(bins) - 1, and the smallest
             bin must be smaller than the minimum of the data.)
             For sparse dataset a tuple with the weight to apply for positive
             (in [0]) and negative class (in [1]). Positive class refers to
             labels with at least one positive value.
             default: None
         bins:
             number of bins to apply to a continuous dataset before calculating
             the probability of classes. Can also be an array of bins or 'auto'
             for on optimized shearch of bins.
             default='auto'
         args:
             arguments specific to the different dataloader that can be used.
         kwargs:
             dictionnary with specific arguments to the dataloader.
    """
    def __init__(self, batch_size,
                       one_hot_encoding=True,
                       output_shape=None,
                       weighting_mode=None,
                       bins='auto',
                       *args,
                       **kwargs):
        self.one_hot_encoding = one_hot_encoding
        self.output_shape = output_shape
        self.weighting_mode = weighting_mode
        self.bins = bins
        self.frame = inspect.currentframe()

        old_shape = StringSeqIntervalDl.predict_label_shape(**kwargs)

        if self.output_shape:
            try:
                new_shape = self.output_shape[1:]
                np.zeros(old_shape).reshape(new_shape)
            except ValueError:
                raise ValueError("""output_shape is incompatible with the other arguments.
                Trying to reshape an array of shape {} in shape {}""".format(old_shape,
                                                                             new_shape))

        if self.one_hot_encoding:
            self.dataset = SeqIntervalDl(*args,
                                         **kwargs)
        else:
            self.dataset = StringSeqIntervalDl(*args,
                                               **kwargs)
        self.batch_size = batch_size
        
        if self.weighting_mode:
            self.weights = Weights(self.dataset,
                                   self.weighting_mode,
                                   self.bins)

    def __call__(self):
        """Returns a generator to train a keras model (yielding inputs and
        outputs)."""
        def generator_function(dataset, batch_size):
            indexes = np.arange(0, len(dataset), 1)
            number_of_batches = len(dataset) // batch_size + 1
            
            while True:
            # reshuffled the train set after an epoch
                np.random.shuffle(indexes)
                
                for num in range(number_of_batches):
                    batch_indexes = indexes[num*batch_size : (num + 1) * batch_size]
                    data = dataset[list(batch_indexes)]
                    inputs = data['inputs']
                    outputs = data['targets']
                    
                    if self.output_shape:
                        if isinstance(outputs, np.ndarray):
                            outputs = outputs.reshape((outputs.shape[0],) +\
                                                      tuple(self.output_shape)[1:])
                        else:
                            outputs[0] = outputs.reshape((outputs[0].shape[0],) +\
                                                          tuple(self.output_shape)[1:])
                    if self.weighting_mode:
                        weights = self.weights.find_weights(outputs)
                        
                        yield inputs, outputs, weights
                    else:
                        yield inputs, outputs
            
        return generator_function(self.dataset, self.batch_size)

    def __len__(self):
        return len(self.dataset) // self.batch_size + 1

    @property
    def command_dict(self):
        return ArgumentsDict(self, called_args='dataset')

    @classmethod
    def default_dict(cls):
        return get_default_args(cls.__init__)
    
    @classmethod
    def predict_sec_input_shape(cls, **input_dict):
        return StringSeqIntervalDl.predict_sec_input_shape(**input_dict)

    @property
    def secondary_input_shape(self):
        try:
            return tuple(self.dataset.secondary_input_shape)
        except AttributeError:
            return tuple(self.dataset.seq_dl.secondary_input)
    
    @classmethod
    def predict_label_shape(cls, **input_dict):
        command_dict = cls.default_dict()
        command_dict.update(input_dict)

        output_shape = StringSeqIntervalDl.predict_label_shape(**command_dict)

        if command_dict['output_shape']:
            try:
                new_shape = command_dict['output_shape'][1:]
                np.zeros(output_shape).reshape(new_shape)
                output_shape = new_shape
            except ValueError:
                raise ValueError("""output_shape is incompatible with the other arguments.
                Try to reshape an array of shape {} in shape {}""".format(output_shape,
                                                                          new_shape))
        return output_shape

    @property
    def label_shape(self):
        command_dict = self.command_dict.as_input()

        if command_dict['output_shape']:
            return tuple(command_dict['output_shape'][1:])
        else:
            return tuple(self.dataset.label_shape)

    @classmethod
    def predict_input_shape(cls, **input_dict):
        command_dict = cls.default_dict()
        command_dict.update(input_dict)
        
        if command_dict['one_hot_encoding']:
            return SeqIntervalDl.predict_input_shape(**command_dict)
        else:
            return None

    @property
    def input_shape(self):
        command_dict = self.command_dict.as_input()

        if command_dict['one_hot_encoding']:
            return tuple(self.dataset.input_shape)
        else:
            return None


class MultiGenerator(object):
    """
    info:
        doc: >
            Class able to yield inputs and targets from several different
            interval readers. Usefull to train on several species or to train
            on both direct and reverse side.
     args:
         batch_size:
             number of example per batch pass to the model.
         dataset_list:
             list of SeqIntervalDl or StringSeqIntervalDl example, the output
             shape need to be the same for all instance.
         inst_per_dataset:
             list of integer, number of example to be taken from each dataset.
             default='all'
         output_shape:
             How to modify the shape of the output (because the initial output
             structure is (batch, length, nb_types, nb_annotation) or (batch,
             nb_types, nb_annotation))
    """
    def __init__(self, batch_size,
                       dataset_list,
                       inst_per_dataset='all',
                       output_shape=None):
        self.dataset_list = dataset_list
        self.batch_size = batch_size
        self.inst_per_dataset = inst_per_dataset
        self.output_shape = output_shape
        self.frame = inspect.currentframe()
        self._verify_dataset_list()

    def __call__(self):
        """Returns a generator to train a keras model (yielding inputs and
        outputs)."""
        indexes = self._get_indexes()
        def generator_function(list_of_dataset, batch_size):
            number_of_batches = len(indexes) // batch_size
            
            while True:
            # reshuffled the train set after an epoch
                np.random.shuffle(indexes)
            
                for num in range(number_of_batches):
                    batch_indexes = indexes[num*batch_size : (num + 1) * batch_size]
                    inputs = None
                    targets = None

                    for dataset_index, dataset in enumerate(list_of_dataset): 
                        sub_batch_indexes = batch_indexes[batch_indexes[:, 0] == dataset_index]
                        data = dataset[sub_batch_indexes[:, 1].tolist()]
                        inputs = self._append_data(inputs, data['inputs'])
                        targets = self._append_data(targets, data['targets'])
                    if self.output_shape:
                        targets = targets.reshape((targets.shape[0],) +\
                                                   tuple(self.output_shape)[1:])
                    yield inputs, targets
            
        return generator_function(self.dataset_list, self.batch_size)

    def _append_data(self, ldata, rdata):
        if isinstance(ldata, list):
            if len(rdata[0]) != 0:
                ldata[0] = np.append(ldata[0], rdata[0], axis=0)
                ldata[1] = np.append(ldata[1], rdata[1], axis=0)
        
        elif isinstance(ldata, np.ndarray):
            if len(rdata) != 0:
                ldata = np.append(ldata, rdata, axis=0)
        else:
            if isinstance(rdata, list):
                if len(rdata[0]) != 0:
                    ldata = rdata
            elif isinstance(rdata, np.ndarray):
                if len(rdata) != 0:
                    ldata = rdata
        return ldata

    def _get_indexes(self):
        if not self.inst_per_dataset == 'all':
            assert len(self.dataset_list) == len(self.inst_per_dataset),\
            """To pass the number of examples to be taken from every dataset,
            the list of dataset and list of number must be of the same length
            """

        indexes = np.zeros((1, 2))
        if self.inst_per_dataset == 'all':
            self.inst_per_dataset = [len(dataset) for dataset in self.dataset_list]
        
        for dataset_index, dataset in enumerate(self.dataset_list):
            length = len(dataset)
            nb_inst = self.inst_per_dataset[dataset_index]
            indexes_ = np.append(np.repeat([[dataset_index]], nb_inst,
                                           axis=0),
                                 np.random.choice(np.arange(length),
                                                  nb_inst,
                                                  replace=False).reshape(nb_inst,
                                                                         1),
                                 axis=1)

            indexes = np.append(indexes, indexes_, axis=0)
        indexes = indexes[1:].astype(int)
        return indexes

    def __len__(self):
        return len(self._get_indexes()) // self.batch_size

    @property
    def command_dict(self):
        argsdict_list = [dataset.command_dict for dataset in self.dataset_list]
        argsdict_list.append(ArgumentsDict(self, kwargs=False))
        return argsdict_list

    def _verify_dataset_list(self):
        double_check = zip(self.dataset_list[:-1],
                           self.dataset_list[1:])

        for d1, d2 in double_check:
            assert type(d1) == type(d2),\
            """All dataset should be of the same class"""

            try:
                assert d1.input_shape == d2.input_shape,\
                """Found incompatible input shape: {} and {}""".format(d1.input_shape,
                                                                       d2.input_shape)
            except NameError:
                pass

            assert d1.secondary_input_shape == d2.secondary_input_shape,\
            """Found incompatible secondary input shape: {} and {}""".\
            format(d1.secondary_input_shape, d2.secondary_input_shape)

            assert d1.label_shape == d2.label_shape,\
            """Found incompatible label shape {} and {}""".format(d1.label_shape,
                                                                  d2.label_shape)

        if self.output_shape:
            try:
                np.zeros(self.output_shape[1:]).reshape(self.dataset_list[0].label_shape)
            except ValueError:
                raise ValueError("""output_shape is incompatible with the other arguments.
                Trying to reshape an array of shape {} in shape {}""".\
                                 format(self.output_shape[1:],
                                        self.dataset_list[0].label_shape))

    @property
    def input_shape(self):
        try:
            return self.dataset_list[0].input_shape
        except NameError:
            return None

    @property
    def secondary_input_shape(self):
        return self.dataset_list[0].secondary_input_shape

    @property
    def label_shape(self):
        if self.output_shape:
            return self.output_shape[1:]
        else:
            return self.dataset_list[0].label_shape 


class PredictionGenerator(object):
    """
    info:
        doc: >
            Takes the command dict of a Generator instance (or a SeqIntervalDl
            or a StringSeqIntervalDl instance) and returns a generator needed
            to predict all along chromosomes.
    
    args:
        batch_size:
            integer
        chrom_size:
            A .sizes file with the size of chromosomes in two columns.
        command_dict:
            An ArgumentsDict instance that described how the model was trained.
        incl_chromosomes:
            list of chromosome to predict on.
        start_stop:
            list of tuple indicating where to start and stop generating data.
            One per included chromosome. default=None
        fasta_file:
            A fasta_file to make prediction on (if None the same as in
            command_dict)
            default=None
        rc:
            Weither or not to predict with reverse complemented DNA sequences.
            default=False
        overlapping:
            int or False, if int the intervals are overalpped with a stride of
            int.
            default=False
    """
    def __init__(self,
                 batch_size,
                 command_dict,
                 chrom_size,
                 incl_chromosomes,
                 start_stop=None,
                 fasta_file=None,
                 rc=False,
                 overlapping=False):
        self.batch_size = batch_size
        self.command_dict = command_dict
        self.chrom_size = chrom_size
        self.start_stop = start_stop
        self.rc = rc
        self.overlapping = overlapping

        if isinstance(incl_chromosomes, list):
            self.incl_chromosomes = incl_chromosomes
        else:
            self.incl_chromosomes = [incl_chromosomes]
        self.detailed_dict = command_dict.get_details()

        if 'keras_dna.sequence.SparseDataset' in self.detailed_dict:
            dataset_dict = self.detailed_dict['keras_dna.sequence.SparseDataset']
            self.window = dataset_dict['length']
            self.tg_window = 1

            if dataset_dict['seq2seq']:
                self.tg_window = self.window

        elif 'keras_dna.sequence.ContinuousDataset' in self.detailed_dict:
            dataset_dict = self.detailed_dict['keras_dna.sequence.ContinuousDataset']
            self.window = dataset_dict['window']

            if dataset_dict['downsampling']:
                self.tg_window = dataset_dict['window']
                self.sampling_len = dataset_dict['window'] // dataset_dict['tg_window']
            else:
                self.tg_window = dataset_dict['tg_window']

        string_dict = deepcopy(self.detailed_dict['keras_dna.sequence.StringSeqIntervalDl'])
        string_dict['annotation_files'] = self.chrom_size
        string_dict['use_strand'] = False
        string_dict['rc'] = self.rc
        if fasta_file:
            string_dict['fasta_file'] = fasta_file

        continuous_dict = {'ignore_targets' : True,
                           'overlapping' : self.overlapping,
                           'window' : self.window,
                           'incl_chromosomes' : self.incl_chromosomes,
                           'start_stop' : self.start_stop,
                           'tg_window' : self.tg_window}

        if 'keras_dna.sequence.SeqIntervalDl' in self.detailed_dict:
            self.input_dict = deepcopy(self.detailed_dict['keras_dna.sequence.SeqIntervalDl'])
            self.input_dict.update(string_dict)
            self.input_dict.update(continuous_dict)
            self.dataset = SeqIntervalDl(**self.input_dict)
        else:
            self.input_dict = string_dict
            self.input_dict.update(continuous_dict)
            self.dataset = StringSeqIntervalDl(**self.input_dict)

    def __len__(self):
        return len(self.dataset) // self.batch_size + 1

    def __call__(self):
        """Returns a generator to train a keras model (yielding inputs and
        outputs)."""
        def generator_function(dataset, batch_size):
            indexes = np.arange(len(dataset))
            number_of_batches = len(dataset) // batch_size + 1

            while True:
                for num in range(number_of_batches):
                    if num == number_of_batches - 1:
                        batch_indexes = indexes[num * batch_size :]
                    else:
                        batch_indexes = indexes[num*batch_size : (num + 1) * batch_size]
                    data = dataset[list(batch_indexes)]
                    yield data['inputs']

        return generator_function(self.dataset, self.batch_size)

    @property
    def index_df(self):
        """
        Returns a DataFrame with the first and last index for every chromosome
        and the position of the begining (included) of the prediction and the
        end (excluded) of the prediction on the chromosome.
        """
        if 'keras_dna.sequence.SeqIntervalDl' in self.detailed_dict:
            df = deepcopy(self.dataset.seq_dl.dataset.df)
        else:
            df = deepcopy(self.dataset.dataset.df)

        if 'keras_dna.sequence.SparseDataset' in self.detailed_dict:
            dataset_dict = self.detailed_dict['keras_dna.sequence.SparseDataset']

            if dataset_dict['seq2seq']:
                df['start'] = df.start.values - self.window // 2
                df['stop'] = df.start.values + (df.last_index.values - df.first_index.values)\
                            * self.window + self.window
            else:
                df['start'] = df.start.values
                df['stop'] = df.start.values + (df.last_index.values - df.first_index.values)\
                            + 1
        elif 'keras_dna.sequence.ContinuousDataset' in self.detailed_dict:
            dataset_dict = self.detailed_dict['keras_dna.sequence.ContinuousDataset']

            if dataset_dict['downsampling']:
                df['start'] = df.start.values - self.window // 2
                df['stop'] = df.start.values + (df.last_index.values - df.first_index.values)\
                            * self.window + self.window
            else:
                tg_window = dataset_dict['tg_window']
                df['start'] = df.start.values - tg_window // 2
                df['stop'] = df.start.values + (df.last_index.values - df.first_index.values)\
                            * tg_window + tg_window
        return df


class VariantEffectGenerator(object):
    """
    info:
        doc: >
            Takes the command dict of a Generator instance (or a SeqIntervalDl
            or a StringSeqIntervalDl instance) and returns a generator needed
            to the effect of individual mutations all along chromosomes.
    
    args:
        chrom_size:
            A .sizes file with the size of chromosomes in two columns.
        command_dict:
            An ArgumentsDict instance that described how the model was trained.
        incl_chromosomes:
            Chromosome to predict on (one by one as it is a long process)
        start_stop:
            list of tuple indicating where to start and stop generating data.
            One per included chromosome. default=None
        fasta_file:
            A fasta_file to make prediction on (if None the same as in
            command_dict)
            default=None
    """
    def __init__(self,
                 command_dict,
                 chrom_size,
                 incl_chromosomes,
                 start_stop=None,
                 fasta_file=None):
        self.command_dict = command_dict
        self.chrom_size = chrom_size
        self.start_stop = start_stop

        if isinstance(incl_chromosomes, list):
            self.incl_chromosomes = incl_chromosomes
        else:
            self.incl_chromosomes = [incl_chromosomes]
        self.detailed_dict = command_dict.get_details()

        assert not 'keras_dna.sequence.SparseDataset' in self.detailed_dict,\
        """Sparse mutation effect is not supported yet"""

        if 'keras_dna.sequence.ContinuousDataset' in self.detailed_dict:
            dataset_dict = self.detailed_dict['keras_dna.sequence.ContinuousDataset']
            self.window = dataset_dict['window']

            if dataset_dict['downsampling']:
                self.tg_window = dataset_dict['window']
                self.sampling_len = dataset_dict['window'] // dataset_dict['tg_window']
                self.tg_downsampling = dataset_dict['tg_window']
            else:
                self.tg_window = dataset_dict['tg_window']
                self.tg_downsampling = None

        string_dict = deepcopy(self.detailed_dict['keras_dna.sequence.StringSeqIntervalDl'])
        string_dict['annotation_files'] = self.chrom_size
        string_dict['use_strand'] = False
        string_dict['rc'] = False
        if fasta_file:
            string_dict['fasta_file'] = fasta_file

        continuous_dict = {'ignore_targets' : True,
                           'overlapping' : False,
                           'window' : self.window,
                           'incl_chromosomes' : self.incl_chromosomes,
                           'start_stop' : self.start_stop,
                           'tg_window' : self.tg_window}

        if 'keras_dna.sequence.SeqIntervalDl' in self.detailed_dict:
            self.input_dict = deepcopy(self.detailed_dict['keras_dna.sequence.SeqIntervalDl'])
            self.input_dict.update(string_dict)
            self.input_dict.update(continuous_dict)
            self.dataset = SeqIntervalDl(**self.input_dict)
        else:
            self.input_dict = string_dict
            self.input_dict.update(continuous_dict)
            self.dataset = StringSeqIntervalDl(**self.input_dict)

    def __len__(self):
        return len(self.dataset) * self.window * 4

    def __call__(self):
        """Returns a generator to train a keras model (yielding inputs and
        outputs)."""
        def generator_function(dataset):
            while True:
                for num in range(len(dataset)):
                    data = dataset[num]['inputs'][0]
                    expanded_dims = None

                    if len(data.shape) == 3:
                        expanded_dims = np.where(np.array(data.shape) == 1)[0][0]

                        if expanded_dims == 1:
                            data = data[:, 0, :]
                        else:
                            data = data[:, :, 0]

                    for i in range(len(data)):
                        nucleotid = np.where(data[i] == 1)[0][0]
                        loc_data = np.tile(data, [4, 1, 1])

                        for j in range(4):
                            if j > 0:
                                loc_data[j, i, nucleotid] = 0
                                loc_data[j, i, (nucleotid + j) % 4] = 1

                        if expanded_dims:
                            loc_data = np.expand_dims(loc_data, expanded_dims + 1)
                        yield loc_data

        return generator_function(self.dataset)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from tensorflow.keras.layers import Reshape, Layer
import tensorflow.keras.backend as K


class Project1D(Layer):
    """
    Layer designed to change the output of a model with several annotation or
    cell_type to a single output.
    """
    def __init__(self,
                 cell_idx,
                 ann_idx,
                 nb_cell_type,
                 nb_annotation_type,
                 **kwargs):
        super(Project1D, self).__init__(**kwargs)
        self.cell_idx = cell_idx
        self.ann_idx = ann_idx
        self.nb_cell_type = nb_cell_type
        self.nb_annotation_type = nb_annotation_type

    def call(self, inputs):
        tensor = self._project(inputs)
        return Reshape((1,))(tensor)

    def compute_output_shape(self, input_shape):
        return (input_shape[0], 1)

    def _project(self, tensor):
        if self.nb_cell_type == 1 and self.nb_annotation_type != 1:
            if len(K.int_shape(tensor)) == 3:
                tensor = tensor[:, self.cell_idx, self.ann_idx]
            elif len(K.int_shape(tensor)) == 2:
                tensor = tensor[:, self.ann_idx]
        
        elif self.nb_cell_type != 1 and self.nb_annotation_type == 1:
            if len(K.int_shape(tensor)) == 3:
                tensor = tensor[:, self.cell_idx, self.ann_idx]
            elif len(K.int_shape(tensor)) == 2:
                tensor = tensor[:, self.cell_idx]
        elif self.nb_cell_type == 1 and self.nb_annotation_type == 1:
            if len(K.int_shape(tensor)) == 3:
                tensor = tensor[:, 0, 0]
            elif len(K.int_shape(tensor)) == 2:
                tensor = tensor[:, 0]
        else:
            tensor = tensor[:, self.cell_idx, self.ann_idx]
        return tensor
    
    
class Project2D(Layer):
    """
    Layer designed to change the output of a model with several annotation or
    cell_type to a single output for continuous annotation.
    """
    def __init__(self,
                 cell_idx,
                 ann_idx,
                 nb_cell_type,
                 nb_annotation_type,
                 **kwargs):
        super(Project2D, self).__init__(**kwargs)
        self.cell_idx = cell_idx
        self.ann_idx = ann_idx
        self.nb_cell_type = nb_cell_type
        self.nb_annotation_type = nb_annotation_type

    def call(self, inputs):
        return self._project(inputs)

    def compute_output_shape(self, input_shape):
        return input_shape[:2]

    def _project(self, tensor):
        if self.nb_types == 1 and self.nb_annotation != 1:
            try:
                tensor = tensor[:, :, self.cell_idx, self.idx]
            except ValueError:
                tensor = tensor[:, :, self.idx]

        elif self.nb_types != 1 and self.nb_annotation == 1:
            try:
                tensor = tensor[:, :, self.cell_idx, self.idx]
            except ValueError:
                tensor = tensor[:, :, self.cell_idx]
        elif self.nb_types == 1 and self.nb_annotation == 1:
            try:
                tensor = tensor[:, :, 0, 0]
            except ValueError:
                try:
                    tensor = tensor[:, :, 0]
                except ValueError:
                    tensor = tensor
        return tensor
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Dec 11 16:33:49 2019

@author: routhier
"""

import tensorflow as tf
from tensorflow.keras import backend as K


class Auc(object):
    """
    info:
        doc: >
            Class usefull to evaluate a sparse model with the AUROC metric on
            a specific cellular type and annotation if the model outputs several
            of them.
    args:
        curve:
            {'ROC', 'PR'} two available metrics.
            default='ROC'
    """
    def __init__(self,
                 curve='ROC'):
        assert curve in ['ROC', 'PR'], \
        """ The two available metrics are 'ROC' and 'PR' """
        self.curve = curve
        
    def metric(self,
               y_pred,
               y_true):
        auc, update_op = tf.metrics.auc(y_pred, y_true, curve=self.curve)
        K.get_session().run(tf.local_variables_initializer())
        
        with tf.control_dependencies([update_op]):
            auc = tf.identity(auc)
        return auc


def correlate(y_pred, y_true, cell_idx, idx, nb_types, nb_annotation):
    y_pred = _project(y_pred, cell_idx, idx, nb_types, nb_annotation)
    y_true = _project(y_true, cell_idx, idx, nb_types, nb_annotation)

    X = y_true - K.mean(y_true)
    Y = y_pred - K.mean(y_pred)

    sigma_XY = K.sum(X*Y)
    sigma_X = K.sqrt(K.sum(X*X))
    sigma_Y = K.sqrt(K.sum(Y*Y))

    return sigma_XY/(sigma_X*sigma_Y + K.epsilon())

def _project(y, cell_idx, idx, nb_types, nb_annotation):
    if nb_types == 1 and nb_annotation != 1:
        if len(K.int_shape(y)) == 4:
            y = y[:, :, cell_idx, idx]
        elif len(K.int_shape(y)) == 3:
            y = y[:, :, idx]

    elif nb_types != 1 and nb_annotation == 1:
        if len(K.int_shape(y)) == 4:
            y = y[:, :, cell_idx, idx]
        elif len(K.int_shape(y)) == 3:
            y = y[:, :, cell_idx]

    elif nb_types == 1 and nb_annotation == 1:
        if len(K.int_shape(y)) == 4:
            y = y[:, :, 0, 0]
        elif len(K.int_shape(y)) == 3:
            y = y[:, :, 0]

    return y



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Nov 13 11:30:24 2020

@author: routhier
"""

import numpy as np
import math
import tensorflow.keras as keras


def get_maximum_first_layer_activations(first_layer_model):
    first_layer, biases = first_layer_model.layers[0].get_weights()
    return np.sum(np.max(np.abs(first_layer), axis=1), axis=0) + biases

def get_maximum_second_layer_activations(first_layers_model):
    max_first_layer = get_maximum_first_layer_activations(first_layers_model)

    second_layer, biases = first_layers_model.layers[-1].get_weights()
    activation = first_layers_model.layers[0].activation
    max_first_layer = activation(max_first_layer).numpy()

    if len(first_layers_model.layers) == 4:
        gamma, beta, mean, std = first_layers_model.layers[2].get_weights()
        std[std < 10e-5] = 1
        max_first_layer = gamma * (max_first_layer - mean) / std + beta

    max_first_layer = np.tile(max_first_layer, [len(second_layer), second_layer.shape[2], 1])
    max_first_layer = np.swapaxes(max_first_layer, 1, 2)

    second_layer[second_layer < 0] = 0

    max_second_layer = max_first_layer * second_layer
    return np.sum(np.sum(max_second_layer, axis=1), axis=0) + biases

def find_pfm_on_batch(sequences,
                      first_layer_model,
                      pool_size,
                      threshold=0.5,
                      layer='first_layer'):
    if layer == 'first_layer':
        max_activation = get_maximum_first_layer_activations(first_layer_model)
    else:
        max_activation = get_maximum_second_layer_activations(first_layer_model)
    seq_shape = np.array(sequences.shape)
    
    if len(seq_shape) == 4:
        axis = np.where(seq_shape == 1)[0][0]
        if axis == 2:
            sequences = sequences[:, :, 0, :]
        else:
            sequences = sequences[:, :, :, 0]
    
    if layer == 'first_layer':
        length = len(first_layer_model.layers[0].get_weights()[0])
    else:
        if len(first_layer_model.layers) == 4:
            conv_idx = 3
        else:
            conv_idx = 2
        length = len(first_layer_model.layers[conv_idx].get_weights()[0]) * pool_size \
                + len(first_layer_model.layers[0].get_weights()[0]) - 1 

    activations = first_layer_model.predict(sequences)
    activations = np.swapaxes(activations, 0, 2)
    activations = np.swapaxes(activations, 1, 2)
    
    max_activation = np.repeat(max_activation,
                               activations.shape[1] * activations.shape[2]).reshape(activations.shape)

    kernel_indexes, seq_indexes, positions = np.where(activations >= threshold * max_activation)
    positions = positions * pool_size
    
    number_activator_seq = [len(np.unique(seq_indexes[kernel_indexes == i]))\
                            for i in range(len(activations))]
    number_activations = [len(seq_indexes[kernel_indexes == i])\
                          for i in range(len(activations))]
    return np.concatenate([extract_kernel_on_batch(kernel, sequences, kernel_indexes, seq_indexes, positions, length) for kernel in range(len(max_activation))], 0),\
np.array(number_activator_seq), np.array(number_activations)

def extract_kernel_on_batch(kernel, sequence, kernel_indexes, seq_indexes, positions, length):
    if len(seq_indexes[kernel_indexes == kernel]) > 0:
        pfm = np.mean(sequence[np.repeat(seq_indexes[kernel_indexes == kernel], length).reshape((len(seq_indexes[kernel_indexes == kernel]), length)),
                     np.concatenate([np.expand_dims(np.arange(pos, pos + length), 0) for pos in positions[kernel_indexes == kernel]], axis=0)], axis=0)
        return np.expand_dims(pfm, 0)
    else:
        return np.zeros((1, length, 4))

def find_pfm(generator,
             first_layer_model,
             pool_size,
             threshold=0.5,
             layer='first_layer'):
    number_of_batches = len(generator)
    new_generator = generator()

    for i in range(number_of_batches):
        if i == 0:
            sequences = next(new_generator)[0]
            nb_seq = len(sequences)
            pfms, nb_act_seqs, nb_acts = find_pfm_on_batch(sequences,
                                                           first_layer_model,
                                                           pool_size,
                                                           threshold,
                                                           layer)
            normalizer = np.ones(pfms.shape)
            normalizer[np.sum(np.sum(pfms, axis=2), axis=1) == 0] = 0

        else:
            sequences = next(new_generator)[0]
            nb_seq += len(sequences)
            pfm, nb_act_seq, nb_act = find_pfm_on_batch(sequences,
                                                        first_layer_model,
                                                        pool_size,
                                                        threshold,
                                                        layer)
            pfms += pfm
            nb_act_seqs += nb_act_seq
            nb_acts += nb_act
            counter = np.ones(pfms.shape)
            counter[np.sum(np.sum(pfm, axis=2), axis=1) == 0] = 0
            normalizer += counter

    return pfms / normalizer, nb_act_seqs / nb_seq, nb_acts / nb_act_seqs

def create_logos(pfm):
    pfm[pfm == 0] = 0.001
    return  pfm * (2 + np.concatenate([np.expand_dims(np.sum(pfm * np.log2(pfm), axis=-1), -1) for i in range(4)], -1))

def get_conv_parameters(model, idx):
    if isinstance(model.layers[idx], keras.layers.Conv1D):
        length_filters, alphabet, nb_filters = model.layers[idx].get_weights()[0].shape
        _, length_seq, alphabet = keras.backend.int_shape(model.input)
        return length_seq, length_filters, alphabet, nb_filters

    elif isinstance(model.layers[idx], keras.layers.Conv2D):
        _, length_seq, dummy, alphabet = keras.backend.int_shape(model.input)
        if dummy == 1:
            length_filters, _, alphabet, nb_filters = model.layers[idx].get_weights()[0].shape
        else:
            length_filters, alphabet, _, nb_filters = model.layers[idx].get_weights()[0].shape
        return length_seq, length_filters, alphabet, nb_filters, dummy
    
def find_conv_index(model):
    layers = model.layers

    conv1d_indexes = []
    conv2d_indexes = []
    maxpool_indexes = []
    batchnorm_indexes = []

    for i, layer in enumerate(layers):
        if isinstance(layer, keras.layers.Conv1D):
            conv1d_indexes.append(i)
        if isinstance(layer, keras.layers.Conv2D):
            conv2d_indexes.append(i)
        if isinstance(layer, keras.layers.BatchNormalization):
            batchnorm_indexes.append(i)

    for i, dico in enumerate(model.get_config()['layers']):
        if dico['class_name'] == 'MaxPooling1D' or\
        dico['class_name'] == 'MaxPooling2D':
            maxpool_indexes.append(i)
    return conv1d_indexes, conv2d_indexes, maxpool_indexes, batchnorm_indexes

def add_conv_layer(model,
                   first_layers_model,
                   idx,
                   dimension,
                   activation,
                   input_shape=True):
    if dimension == 'Conv1D':
        length_seq, length_filters, alphabet, nb_filters = \
        get_conv_parameters(model, idx)
    elif dimension == 'Conv2D':
        length_seq, length_filters, alphabet, nb_filters, dummy = \
        get_conv_parameters(model, idx)
        
    if input_shape:
        input_shape = (length_seq, alphabet)
        first_layers_model.add(keras.layers.Conv1D(filters=nb_filters,
                                               kernel_size=(length_filters,),
                                               input_shape=input_shape,
                                               padding='valid',
                                               activation=activation))
    else:
        first_layers_model.add(keras.layers.Conv1D(filters=nb_filters,
                                               kernel_size=(length_filters,),
                                               padding='valid',
                                               activation=activation))

    if dimension == 'Conv1D':
        first_layers_model.layers[-1].set_weights(model.layers[idx].get_weights())
    elif dimension == 'Conv2D':
        weights, biases = model.layers[idx].get_weights()
        if dummy == 1:
            first_layers_model.layers[-1].set_weights([weights[:, 0, :, :],
                                                       biases])
        else:
            first_layers_model.layers[-1].set_weights([weights[:, :, 0, :],
                                                       biases])

def create_first_layer_model(model, layer='first_layer'):
    first_layers_model = keras.models.Sequential()
    conv1d_indexes, conv2d_indexes, maxpool_indexes, batchnorm_indexes = find_conv_index(model)

    if conv1d_indexes:
        if layer == 'first_layer':
            activation='linear'
        else:
            activation = model.layers[conv1d_indexes[0]].activation
        add_conv_layer(model,
                       first_layers_model,
                       conv1d_indexes[0],
                       'Conv1D',
                       activation)
    elif conv2d_indexes:
        if layer == 'first_layer':
            activation='linear'
        else:
            activation = model.layers[conv2d_indexes[0]].activation
        add_conv_layer(model,
                       first_layers_model,
                       conv2d_indexes[0],
                       'Conv2D',
                       activation)
    else:
        raise ValueError('''The model must contain a convolutional layer''')
    
    if layer == 'first_layer':
        return (first_layers_model, 1)
    else:
        if maxpool_indexes:
            pool_size = model.get_config()['layers'][maxpool_indexes[0]]\
            ['config']['pool_size'][0]
        else:
            pool_size = 1

        first_layers_model.add(keras.layers.MaxPooling1D(pool_size, padding='valid'))

        if batchnorm_indexes:
            idx = batchnorm_indexes[0]
            first_layers_model.add(keras.layers.BatchNormalization())
            first_layers_model.layers[-1].set_weights(model.layers[idx].get_weights())
        
        if conv1d_indexes:
            add_conv_layer(model, first_layers_model, conv1d_indexes[1],
                           'Conv1D', 'linear', False)
        else:
            add_conv_layer(model, first_layers_model, conv2d_indexes[1],
                           'Conv2D', 'linear', False)
        return (first_layers_model, pool_size)

def export_to_meme(logos, name, alphabet):
    with open(name, 'w') as f:
        f.write('MEME version 4\n')
        f.write('ALPHABET= ' + alphabet + '\n')
        f.write('strand= +\n')

        for i in range(len(logos)):
            f.write('\n')
            f.write('MOTIF FILTER_' + str(i) + '\n')
            f.write('letter-probability matrix: alength= 4 w= ' + str(len(logos[i])) + ' nsites= ' + str(len(logos[i])) + ' E= 1.0e-001\n')

            for line in logos[i]:
                for num in line:
                    if num == 0:
                        f.write('0.0\t')
                    elif math.isnan(num):
                        f.write('0.25\t')
                    else:
                        f.write(str(num)+'\t')

                f.write('\n')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov  5 16:40:31 2019

@author: routhier
"""

import pyBigWig
import re
import numpy as np


class Normalizer(object):
    """
    Normalize the data by standard procedure. It takes the wanted procedure and
    a bbi file as input. When called with a sequence extracted from the bbi
    file, the sequence is normalized.
        
    args:
        normalization:
            {zscore, max, perctrim, min_max, None}, name of the procedure to be
            applied.
        bbi_file:
            The file on which the data will be taken and then normalized.
        threshold:
            The pourcentage above which the distribution will be trimed.
    """ 
    def __init__(self,
                 normalization,
                 bbi_file,
                 threshold=99):
        self.normalization = normalization
        self.bbi_file = bbi_file
        self.bw = pyBigWig.open(bbi_file)
        self.threshold = threshold
        self.chrom_size = dict()

        for name, size in self.bw.chroms().items():
            if re.search('^(chr)?(\d+)?([XVI]+)?$', name):
                self.chrom_size[name] = size

        assert self.bw.isBigWig() or self.bw.isBigBed(),\
        """The file passed should be either a bigWig or a bigBed"""
        
        if self.normalization == 'zscore':
            self.mean = np.sum([size * self.bw.stats(name)[0]\
                                for name, size in self.chrom_size.items()])
            self.mean /= np.sum([size for name in self.chrom_size.values()])
            self.std = np.sum([size * self.bw.stats(name, type='std')[0] for name,
                               size in self.chrom_size.items()])
            self.std /= np.sum([size for name in self.chrom_size.values()])

        if self.normalization == 'max':
            self.max = np.max([self.bw.stats(name, type='max')[0]\
                               for name in self.chrom_size.keys()])

        if self.normalization == 'perctrim':
            sampled_values = get_sample(self.bw, self.chrom_size)
            self.limit = np.percentile(sampled_values, self.threshold)
            
        if self.normalization == 'min_max':
            self.max = np.max([self.bw.stats(name, type='max')[0]\
                               for name in self.chrom_size.keys()])
            self.min = np.min([self.bw.stats(name, type='min')[0]\
                               for name in self.chrom_size.keys()])
        self.bw.close()
    
    def __call__(self, seq):
        return self._normalize(seq)
        
    def _normalize(self, seq):
        if self.normalization == 'zscore':
            return (seq - self.mean) / self.std
        elif self.normalization == 'max':
            return (seq / self.max)
        elif self.normalization == 'perctrim':
            seq[seq > self.limit] = self.limit
            return seq
        elif self.normalization is None:
            return seq
        elif self.normalization == 'logtransform':
            return np.log(seq + 1)
        elif self.normalization == 'min_max':
            return (seq - self.min) / (self.max - self.min)
        
        
class BiNormalizer(Normalizer):
    
    def __init__(self,
                 normalization,
                 bbi_file,
                 threshold=99):
        assert len(normalization) == 2,\
        """BiNormalizer can only handle two successive normalization process"""
        self.normalization = normalization[0]
        self.threshold = threshold
        self.first_normalizer = Normalizer(normalization[1],
                                           bbi_file,
                                           self.threshold)
        self.bw = pyBigWig.open(bbi_file)
        self.sample = self.first_normalizer(get_sample(self.bw,
                                                       self.first_normalizer.chrom_size))
        self.bw.close()
        if self.normalization == 'zscore':
            self.mean = np.mean(self.sample)
            self.std = np.std(self.sample)

        if self.normalization == 'max':
            self.max = np.max(self.sample)

        if self.normalization == 'perctrim':
            self.limit = np.percentile(self.sample, self.threshold)
            
        if self.normalization == 'min_max':
            self.max = np.max(self.sample)
            self.min = np.min(self.sample)

    def __call__(self, seq):
        seq = self.first_normalizer(seq)
        return self._normalize(seq)


def get_sample(bw, chrom_size, sampling_len=1000):
    """Returns a sample of the values of the bbi_file as a numpy array"""
    sampled_values = []
    for name, size in chrom_size.items():
        sampling = np.random.randint(0, size - 1, sampling_len)
        sampled_values.append([bw.values(name, sample, sample + 1)[0]\
                               for sample in sampling])
    
    sampled_values =  np.array(sampled_values).reshape((len(sampled_values)*sampling_len,))
    return sampled_values[np.isfinite(sampled_values)]


class Weights(object):
    """
    info:
        doc: >
            Usefull to add weights to train a model. It calculate the probability
            of every labels that a generator will generate and can leads
            batches of weighting arrays corresponding to batches of labels.
     args:
        dataset:
            the dataset of the generator.
        weighting_mode:
            {None, 'balanced', tuple(weights, bins), tuple(weight_pos, weight_neg)}
            the methodology to set the weights.
            For continuous dataset a tuple with the weights to apply and the
            bins can be parsed (len(weights) == len(bins) - 1, and the smallest
            bin must be smaller than the minimum of the data.)
            For sparse dataset a tuple with the weight to apply for positive
            (in [0]) and negative class (in [1]). Positive class refers to
            labels with at least one positive value.
            default: None
        bins:
            number of bins to apply to a continuous dataset before calculating
            the probability of classes. Can also be an array of bins or 'auto'
            for on optimized shearch of bins.
            default='auto'
    """
    def __init__(self,
                 dataset,
                 weighting_mode=None,
                 bins='auto'):
        self.dataset = dataset
        self.command_dict = dataset.command_dict.get_details()

        if 'keras_dna.sequence.SparseDataset' in self.command_dict:
            try:
                df = dataset.dataset.df
            except AttributeError:
                df = dataset.seq_dl.dataset.df

            if weighting_mode == 'balanced':
                nb_neg = len(df[df.type ==  0])
                nb_pos = len(df[df.type != 0])
                
                self.value_positive = (nb_neg + nb_pos) / float(nb_pos * 2)
                self.value_negative = (nb_neg + nb_pos) / float(nb_neg * 2)

            if isinstance(weighting_mode, tuple):
                self.value_positive = weighting_mode[0]
                self.value_negative = weighting_mode[1]

        elif 'keras_dna.sequence.ContinuousDataset' in self.command_dict:
            try:
                chrom_size = dataset.dataset.chrom_size
                norm_dico = dataset.dataset.extractor.norm_dico
                self.tg_window = dataset.dataset.tg_window
            except AttributeError:
                chrom_size = dataset.seq_dl.dataset.chrom_size
                norm_dico = dataset.seq_dl.dataset.extractor.norm_dico
                self.tg_window = dataset.seq_dl.dataset.tg_window

            samples = list()
            annotation_files = self.command_dict['keras_dna.sequence.ContinuousDataset']['annotation_files']

            if not isinstance(annotation_files, list):
                annotation_files = [annotation_files]

            for file in annotation_files:
                bw = pyBigWig.open(file)
                samples.append(norm_dico[file](get_sample(bw, chrom_size)))
                bw.close()

            if weighting_mode == 'balanced':
                probas = list()
                self.list_bins = list()

                for sample in samples:
                    proba, bins = self._get_proba(sample, bins)
                    proba[proba == 0] = np.min(proba[proba > 0])
                    probas.append(proba)
                    self.list_bins.append(bins)

                self.weights = [1 / proba for proba in probas]
                self.weights = [weight / float(len(bins) - 1) for weight, bins\
                                in zip(self.weights, self.list_bins)]

            if isinstance(weighting_mode, tuple) or\
            isinstance(weighting_mode, list):
                try:
                    assert isinstance(weighting_mode[0], list)\
                    and isinstance(weighting_mode[1], list),\
                    """Weighting_mode should be 'balanced' or a tuple of list
                    or a tuple of numpy arrays"""
                except AssertionError:
                    assert isinstance(weighting_mode[0], np.ndarray)\
                    and isinstance(weighting_mode[1], np.ndarray),\
                    """Weighting_mode should be 'balanced' or a tuple of list
                    or a tuple of numpy arrays"""

                assert len(weighting_mode[0]) == len(annotation_files)\
                and len(weighting_mode[1]) == len(annotation_files),\
                """Weights and bins must be parsed for every annotation_files"""

                assert len(weighting_mode[0][0]) == len(weighting_mode[1][0]) + 1,\
                """len(bins) must be equal to len(weights) + 1 !"""

                self.weights = [weight for weight in weighting_mode[1]]
                self.list_bins = [bins for bins in weighting_mode[0]]

    def _get_proba(self, array, bins):
        counts, values = np.histogram(array,
                                      bins,
                                      range=(min(array) - 0.001, max(array)),
                                      density=True)
        return counts * (values[1] - values[0]), values

    def find_weights(self, seq):
        if 'keras_dna.sequence.SparseDataset' in self.command_dict:
            outputs = np.zeros((len(seq)))
            outputs += self.value_negative
            outputs[np.where(seq == 1)[0]] = self.value_positive
            return outputs
            
        elif 'keras_dna.sequence.ContinuousDataset' in self.command_dict:
            temporal = True
            if len(seq.shape) == 4:
                seq = seq.reshape((seq.shape[0],
                                   seq.shape[1],
                                   seq.shape[2] * seq.shape[3]))
            elif len(seq.shape) == 2:
                if seq.shape[1] == len(self.list_bins):
                    temporal = False
                    seq = np.expand_dims(seq, 1)
                elif seq.shape[1] == self.tg_window:
                    seq = np.expand_dims(seq, 2)
                else:
                    raise(ValueError("""The label are reshaped in a manner that
                    is not supported by 'balanced' keyword"""))

            outputs = np.zeros(seq.shape)

            for i in range(seq.shape[2]):
                digitized = np.digitize(seq[:, :, i],
                                        self.list_bins[i],
                                        right=True)
                digitized[digitized >= len(self.list_bins[i])] = len(self.list_bins[i]) - 1
                outputs[:, :, i] = np.asarray(self.weights[i])[digitized - 1]

            if temporal:
                return np.mean(outputs, axis=2)
            else:
                return np.mean(outputs, axis=2)[:, 0]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 26 11:30:24 2019

@author: routhier
"""
import numpy as np
import pybedtools
import pandas as pd
import pyBigWig
import os
import inspect
from distutils.spawn import find_executable

def get_default_args(func):
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }

def bed_to_df(bedfile, annotation_list):
    bedfile = pybedtools.BedTool(bedfile)
    
    assert len(annotation_list) == 1, \
    """A .bed file can only display the position for one type of 
    annotation."""
    
    chroms, starts, stops, labels = [], [], [], []
    strands = None

    if hasattr(bedfile[0], 'strand'):
        if bedfile[0].strand in ['+', '-']:
            strands = []

    for bed in bedfile:
        chroms.append(bed.chrom)
        starts.append(bed.start)
        stops.append(bed.stop)
        labels.append(annotation_list[0])

        if hasattr(bed, 'strand'):
            if bed.strand in ['+', '-']:
                strands.append(bed.strand)
    
    if strands:
        df = pd.DataFrame({'chrom': chroms,
                           'start': starts,
                           'stop': stops,
                           'label': labels,
                           'strand': strands})
    else:
        df = pd.DataFrame({'chrom': chroms,
                           'start': starts,
                           'stop': stops,
                           'label': labels})
    return df

def gff_to_df(gff_file, annotation_list):
    gff_file = pybedtools.BedTool(gff_file)
    
    chroms, starts, stops, labels = [], [], [], []
    if not gff_file[1].strand == '.':
        strands = []
    else:
        strands=None
    
    gff_file = gff_file.filter(lambda x : x.fields[2] in annotation_list)
    
    for gff in gff_file:
        chroms.append(gff.chrom)
        starts.append(gff.start)
        stops.append(gff.stop)
        labels.append(gff.fields[2])
        
        if isinstance(strands, list):
            strands.append(gff.strand)
    
    if isinstance(strands, list):
        df = pd.DataFrame({'chrom': chroms,
                           'start': starts,
                           'stop': stops,
                           'label': labels,
                           'strand': strands})
    else:
        df = pd.DataFrame({'chrom': chroms,
                           'start': starts,
                           'stop': stops,
                           'label': labels})
    return df

def bedGraph_to_df(bedGraph, chrom_size, path=None):
    assert find_executable('bedGraphToBigWig')\
    or find_executable(os.path.join(path, 'bedGraphToBigWig')),\
    '''bedGraphToBigWig is not available to use wig files.
    If you are using a conda environment you can download it by running :

    conda install -c bioconda ucsc-bedgraphtobigwig.

    Otherwise run keras_dna.get_ucsc in a terminal to download it from UCSC.
    After that you will need to specify the path to the directory where you
    install the script with the keyword path_to_ucsc.'''
    assert os.path.exists(bedGraph),\
    """The bedGraph file {} does not exists""".format(bedGraph)
    if path:
        output = os.system(path + '/bedGraphToBigWig ' + bedGraph + \
              ' ' + chrom_size + ' ' + bedGraph[:-8] + 'bw')
    else:
         output = os.system('bedGraphToBigWig ' + bedGraph + \
                  ' ' + chrom_size + ' ' + bedGraph[:-8] + 'bw')
    
    if output != 0:
        print("An error has occured. Returned os error code {}".format(output))

def wig_to_df(wig, chrom_size, path=None):
    assert find_executable('wigToBigWig')\
    or find_executable(os.path.join(path, 'wigToBigWig')),\
    '''wigToBigWig is not available to use wig files.
    If you are using a conda environment you can download it by running :

    conda install -c bioconda ucsc-wigtobigwig.

    Otherwise run keras_dna.get_ucsc in a terminal to download it from UCSC.
    After that you will need to specify the path to the directory where you
    install the script with the keyword path_to_ucsc.'''
    assert os.path.exists(wig),\
    """The wig file {} does not exists""".format(wig)
    if path:
         output = os.system(path + '/wigToBigWig ' + wig + \
              ' ' + chrom_size + ' ' + wig[:-3] + 'bw')
    else:
         output = os.system('wigToBigWig ' + wig + \
                  ' ' + chrom_size + ' ' + wig[:-3] + 'bw')
    
    if output != 0:
        print("An error has occured. Returned os error code {}".format(output))

def bbi_extractor(interval, bbi_files, final_dummy_axis=False):
    """
        Function used to extract the coverage of the given bbi file at the
        position of the interval.
    """   
    if not isinstance(bbi_files, list):
        bbi_files = [bbi_files]
    
    seq = list()
    for bbi_file in bbi_files:
        bw = pyBigWig.open(bbi_file)
        seq.append(np.array(bw.values(interval.chrom,
                                      interval.start,
                                      interval.stop)))
    seq =  np.array(seq).T
    
    if final_dummy_axis:
        seq = seq.reshape((seq.shape[0],
                           seq.shape[1],
                           1))
    return seq

def reverse_complement_fa(seq):
    seq = seq[::-1]
    seq = seq.replace('A', '1')
    seq = seq.replace('C', '2')
    seq = seq.replace('T', 'A')
    seq = seq.replace('G', 'C')
    seq = seq.replace('1', 'T')
    seq = seq.replace('2', 'G')
    
    return seq

def reverse_complement(seqs, labels, bbi_seqs=None):
    for i in range(len(seqs)):
        seqs[i] = reverse_complement_fa(seqs[i])
    
    if isinstance(labels, np.ndarray):
        labels = labels[:, ::-1]
    
    if bbi_seqs is not None:
        bbi_seqs = bbi_seqs[:, ::-1]
    
        return seqs, bbi_seqs, labels
    else:
        return seqs, labels

def rolling_window(array, window=(0,), asteps=None, wsteps=None, axes=None, toend=True):  
    """ 
        Take a numpy array and return a view of this array after applying a rolling window.

        This takes a numpy and cut it in several pieces with the size, the stride and the 
        axes asked as needed. You may want to use it in order to create a set of sequences
        from an array.

        :param array: The array we want to cut
        :param window: The length of the window
        :param asteps: The stride between two window applied
        :param wsteps: The stride whitin the window
        :param axes: The axe on which to apply the rolling window
        :param toend: Weither or not to finish the cut
        :type array: numpy array
        :type window: int or tuple
        :type asteps: int or tuple
        :type wsteps: int or tuple
        :type axes: int
        :type toend: boolean
        :return: The view of the array
        :rtype: numpy array

        :Example:

        >>> a = numpy.array([0,1,2,3,4,5])
        >>> rolling_window(a, window = 2, asteps = 2, wsteps = None)
        array([[0,1],
               [2,3],
               [4,5]])
        >>> rolling_window(a, window = 2, asteps = None, wsteps = 2)
        array([[0,2],
               [1,3],
               [2,4]
               [3,5]])
        >>> rolling_window(a, window = 5, asteps = 2, wsteps = None)
        array([[0,1,2,3,4]])

        .. warning:: Be carreful about the combination of window, wsteps and asteps that may raise 
                     ValueError. This function forces the window to be of the asked size and thus 
                     may stop the application of the window before the end.
    """        

    array = np.asarray(array)
    orig_shape = np.asarray(array.shape)
    window = np.atleast_1d(window).astype(int) # maybe crude to cast to int...

    if axes is not None:
        axes = np.atleast_1d(axes)
        w = np.zeros(array.ndim, dtype=int)
        for axis, size in zip(axes, window):
            w[axis] = size
        window = w

    # Check if window is legal:
    if window.ndim > 1:
        raise ValueError("`window` must be one-dimensional.")
    if np.any(window < 0):
        raise ValueError("All elements of `window` must be larger than 1.")
    if len(array.shape) < len(window):
        raise ValueError("`window` length must be less or equal `array` dimension.") 

    _asteps = np.ones_like(orig_shape)
    if asteps is not None:
        asteps = np.atleast_1d(asteps)
        if asteps.ndim != 1:
            raise ValueError("`asteps` must be either a scalar or one dimensional.")
        if len(asteps) > array.ndim:
            raise ValueError("`asteps` cannot be longer then the `array` dimension.")
        # does not enforce alignment, so that steps can be same as window too.
        _asteps[-len(asteps):] = asteps
        
        if np.any(asteps < 1):
             raise ValueError("All elements of `asteps` must be larger then 1.")
    asteps = _asteps

    _wsteps = np.ones_like(window)
    if wsteps is not None:
        wsteps = np.atleast_1d(wsteps)
        if wsteps.shape != window.shape:
            raise ValueError("`wsteps` must have the same shape as `window`.")
        if np.any(wsteps < 0):
             raise ValueError("All elements of `wsteps` must be larger then 0.")

        _wsteps[:] = wsteps
        _wsteps[window == 0] = 1 # make sure that steps are 1 for non-existing dims.
    wsteps = _wsteps

    # Check that the window would not be larger than the original:
    if np.any(orig_shape[-len(window):] < window * wsteps):
        raise ValueError("`window` * `wsteps` larger then `array` in at least one dimension.")

    new_shape = orig_shape # just renaming...

    # For calculating the new shape 0s must act like 1s:
    _window = window.copy()
    _window[_window==0] = 1

    new_shape[-len(window):] += wsteps - _window * wsteps
    new_shape = (new_shape + asteps - 1) // asteps
    # make sure the new_shape is at least 1 in any \"old\" dimension (ie. steps
    # is (too) large, but we do not care.
    new_shape[new_shape < 1] = 1
    shape = new_shape

    strides = np.asarray(array.strides)
    strides *= asteps
    new_strides = array.strides[-len(window):] * wsteps
    
    # The full new shape and strides:
    if toend:
        new_shape = np.concatenate((shape, window))
        new_strides = np.concatenate((strides, new_strides))
    else:
        _ = np.zeros_like(shape)
        _[-len(window):] = window
        _window = _.copy()
        _[-len(window):] = new_strides
        _new_strides = _
        
        new_shape = np.zeros(len(shape)*2, dtype=int)
        new_strides = np.zeros(len(shape)*2, dtypenucleotid=int)
        
        new_shape[::2] = shape
        new_strides[::2] = strides
        new_shape[1::2] = _window
        new_strides[1::2] = _new_strides
    
    new_strides = new_strides[new_shape != 0]
    new_shape = new_shape[new_shape != 0]
    
    return np.lib.stride_tricks.as_strided(array, shape=new_shape, strides=new_strides)


class ProgressBar:
    '''
    Progress bar
    '''
    def __init__ (self, valmax, maxbar, title, valmin=0):
        if valmax == 0:  valmax = 1
        if maxbar > 200: maxbar = 200
        self.valmax = valmax - valmin
        self.valmin = valmin
        self.maxbar = maxbar
        self.title  = title
    
    def update(self, val):
        import sys
        # process
        val = val - self.valmin
        perc  = round((float(val) / float(self.valmax)) * 100)
        scale = 100.0 / float(self.maxbar)
        bar   = int(perc / scale)
  
        # render 
        out = '\r{}/{} [{}{}{}] {}'.format(int(val),
                                           int(self.valmax),
                                           '=' * (bar - 1),
                                           '>' * (1 - int(perc / 100)),
                                           '.' * (self.maxbar - bar),
                                           self.title)
        sys.stdout.write(out)
        sys.stdout.flush()
        

class ArgumentsDict(object):
    """
    Return dictionnary with the arguments passed to the several class necessary
    to build a generator.
    
    args:
        instance:
            An instance of a class used in the process of building a generator.
        called_args:
            When the instance calls another class able to return an ArgumentsDict,
            called_args is the corresponding argument.
        kwargs:
            If the instance accept kwargs to be passed.
    returns:
        An ArgumentsDict instance with the arguments passed to the class to
        construct a generator.
    """
    def __init__(self,
                 instance,
                 called_args=None,
                 kwargs=True):
        assert hasattr(instance, 'frame'),\
        """To access the arguments dict the class must give access to the
        current frame in the __init__ function with the arguments 'frame' """
        self.frame = instance.frame
        self.instance = instance
        
        if called_args:
            assert hasattr(self.instance, called_args),\
            "The instance must own the arguments {}".format(called_args)
            args_instance = getattr(self.instance, called_args)
            assert hasattr(args_instance, 'command_dict'),\
            """The class called with called_args must be able to return an
            ArgumentsDict instance with 'command_dict'"""
            
        self.called_args = called_args
        self.kwargs = kwargs
    
    def __call__(self):
        """
        Returns a dictionnary with the arguments passed to the instance
        separated in two fields, the arguments corresponding to the class and
        kwargs given to other object used by the instance.
        """
        args, _, _, values = inspect.getargvalues(self.frame)
        dico = {i : values[i] for i in args[1:]}            
        
        if self.kwargs:
            return {'args' : dico, 'kwargs' : values['kwargs']}
        else:
            return dico
        
    def as_input(self):
        """
        Returns a dictionnary with the arguments passed to construct the instan
        ce. Can be used as input to reconstruct the same instance.
        """
        args, _, _, values = inspect.getargvalues(self.frame)
        dico = {i : values[i] for i in args[1:]}            
        
        if self.kwargs:
            dico.update(values['kwargs'])
        return dico
    
    def get_details(self):
        """
        Detailed version of the dictionnary with all the class used in the proc
        -ess and their arguments.
        """
        if self.called_args:
            args_instance = getattr(self.instance, self.called_args)
            args_dico = args_instance.command_dict.get_details()
            args_dico.update({str(type(self.instance))[8 : -2] : self.__call__()['args']})
            return args_dico
        else:
            args_dico = self.as_input()
            if hasattr(self.instance, 'length'):
                args_dico.update({'length' : self.instance.length})
            return {str(type(self.instance))[8 : -2] : args_dico}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun 17 10:17:41 2019

@author: routhier
"""

import pandas as pd
import numpy as np
import pybedtools
import random
import warnings
import pyBigWig
import inspect
import sys


from kipoiseq.extractors import FastaStringExtractor
from kipoiseq.transforms import ReorderedOneHot
from kipoiseq.transforms.functional import fixed_len
from kipoiseq.utils import DNA


from . import utils
from .extractors import bbi_extractor


class SparseDataset(object):
    """
    info:
        docs: >
            Reads the positions corresponding to some annotations in a file
            dedicated to store sparse annotation (gff, gtf, bed) and return a
            pybedtool interval corresponding to every annitation as long as a
            label for every interval.
    args:
        annotation_files:
            list of file with annotations (one file per cellular type for
            example, several annotation per file is possible)
        annotation_list:
            list of annotation to be taken into account (name of those
            annotation in the files)
        predict:
            {'all', 'start', 'stop'} weither to predict the annotation, its
            start or its end, default='all'
        seq_len:
            {'MAXLEN', int, 'real'} the length of the intervals. If MAXLEN then
            the length will be the maximal annotation length in the files, real
            mean that the row intervals are returned.
            default='MAXLEN'
        data_augmentation:
            boolean, if true return all the window of the given length where an
            annotation fit entirely in (false = one window per annotation),
            default=False
        seq2seq:
            boolean, if true the label will be of the length of the input
            sequence with 1 where the annotations are in this sequence,
            default=False
        define_positive:
            {'match_all', 'match_any'} if not seq2seq a sequence will be
            considered positive to an annotation if it matches all or any of an
            annotation instance, default='match_all'.
        num_chr:
            if specified, 'chr' in the chromosome name will be dropped,
            default=False
        incl_chromosomes:
            exclusive list of chromosome names to include in the final dataset.
            if not None, only these will be present in the dataset,
            default=None
        excl_chromosomes:
            list of chromosome names to omit from the dataset. default=None
        ignore_targets: 
            if True, target variables are ignored, default=False
        negative_ratio:
            'all' or int, ratio of negative example compared to positive,
            'all' means that all the negative example are returned.
            default=1
        negative_type:
            {'real', 'random', bedfile, None} if real the negative example will be
            taken from sequences far enough from any annotation example. If None this
            function will return only positive example, 'random' will return 
            interval of length 0. Pass a bedfile to precisely set the negative examples.
            default='real'
    """
    def __init__(self, annotation_files,
                       annotation_list,
                       predict='all',
                       seq_len='MAXLEN',
                       data_augmentation=False,
                       seq2seq=False,
                       defined_positive='match_all',
                       num_chr=False,
                       incl_chromosomes=None,
                       excl_chromosomes=None,
                       ignore_targets=False,
                       negative_ratio=1,
                       negative_type='real'):
        self.annotation_files = annotation_files
        self.annotation_list = annotation_list
        self.predict = predict
        self.seq_len = seq_len
        self.data_augmentation = data_augmentation
        self.seq2seq = seq2seq
        self.defined_positive = defined_positive
        self.num_chr = num_chr
        self.incl_chromosomes = incl_chromosomes
        self.excl_chromosomes = excl_chromosomes
        self.ignore_targets = ignore_targets
        self.negative_ratio = negative_ratio
        self.negative_type = negative_type
        self.frame = inspect.currentframe()

        assert not (self.seq_len == 'real' and self.data_augmentation), \
        '''Returning the real position of the annotation is not compatible with
        data_augmentation'''

        if not isinstance(self.annotation_files, list):
            self.annotation_files = [self.annotation_files]

        df_ann_list = list()

        for annotation_file in self.annotation_files:
            if annotation_file.endswith('.bed'):
                df_ann_list.append(utils.bed_to_df(annotation_file,
                                                   self.annotation_list))
            if annotation_file.endswith(('.gff', 'gff3', 'gtf')):
                df_ann_list.append(utils.gff_to_df(annotation_file,
                                                   self.annotation_list))

        self.ann_df = self._multi_cellular_type(df_ann_list)
        self._binarize_label()
        
        if self.num_chr and self.ann_df.iloc[0][0].startswith("chr"):
            self.ann_df['chrom'] = self.ann_df['chrom'].str.replace("^chr", "")
        if not self.num_chr and not self.ann_df.iloc[0][0].startswith("chr"):
            self.ann_df['chrom'] = "chr" + self.ann_df['chrom']
        
        # omit data outside chromosomes
        if incl_chromosomes is not None:
            self.ann_df = self.ann_df[self.ann_df.chrom.isin(incl_chromosomes)]
        if excl_chromosomes is not None:
            self.ann_df = self.ann_df[~self.ann_df.chrom.isin(excl_chromosomes)]
        
        if not self.predict == 'all':
            self._restrict()
        
        if self.seq_len == 'MAXLEN' or self.seq_len == 'real':
            self.length = self._find_maxlen()
        elif isinstance(self.seq_len, int):
            self.length = self.seq_len
        else:
            raise NameError('seq_len should be "MAXLEN", "real" or an integer')

        if self.seq2seq:
            self.defined_positive = 'match_any'

        self.df = self._get_dataframe()

        if not self.ignore_targets:
            self.labels = self._get_labels()

        if self.negative_type == 'random':
            assert isinstance(self.negative_ratio, int), \
            'To use random negative sequence negative_ratio must be an integer'
            self._random_negative_class()

        elif self.negative_type == 'real':
            neg_df, neg_label = self._negative_class()
            self.df = self.df.append(neg_df)

            if not self.ignore_targets:
                self.labels = np.append(self.labels, neg_label, axis=0)

        elif not self.negative_type:
            pass

        elif self.negative_type.endswith('.bed'):
            neg_df, neg_label = self._negative_bed()
            self.df = self.df.append(neg_df)

            if not self.ignore_targets:
                self.labels = np.append(self.labels, neg_label, axis=0)

    @classmethod
    def default_dict(cls):
        return utils.get_default_args(cls.__init__)

    @classmethod
    def predict_label_shape(cls, **input_dict):
        command_dict = cls.default_dict()
        command_dict.update(input_dict)

        if command_dict['ignore_targets']:
            return None

        assert 'annotation_files' in command_dict,\
        """To create an instance passing annotation_files is required"""

        assert 'annotation_list' in command_dict,\
        """To create an instance passing annotation_list is required"""

        if command_dict['seq2seq']:
            assert not isinstance(command_dict['seq_len'], str),\
            """The label shape can only anticipated if seq_len is an integer"""

        if isinstance(command_dict['annotation_files'], str):
            command_dict['annotation_files'] = [command_dict['annotation_files']]
        
        if command_dict['seq2seq']:
            return (command_dict['seq_len'],
                    len(command_dict['annotation_files']),
                    len(command_dict['annotation_list']))
        else:
            return (len(command_dict['annotation_files']),
                    len(command_dict['annotation_list']))

    @property
    def label_shape(self):
        command_dict = self.command_dict.as_input()
        command_dict['seq_len'] = self.length
        return self.predict_label_shape(**command_dict)

    def __getitem__(self, idx):
        """Returns (pybedtools.Interval, labels)"""
        if not isinstance(idx, list):
            idx = [idx]
        row = self.df.iloc[idx]
        
        if self.ignore_targets:
            labels = {}
        else:
            labels = self.labels[idx]
            index = []

        intervals = list()
        
        if 'strand' in self.df.columns:
            for i in range(len(idx)):
                row_ = row.iloc[i]
                try:
                    intervals.append(pybedtools.create_interval_from_list([row_.chrom,
                                                                           int(row_.start),
                                                                           int(row_.stop),
                                                                           '.', '.',
                                                                           row_.strand]))
                    if not self.ignore_targets:
                        index.append(i)

                except OverflowError:
                    warnings.warn("""Some of the input sequence were out of range
                                  and have been removed""")

        else:
            for i in range(len(idx)):
                row_ = row.iloc[i]
                try:
                    intervals.append(pybedtools.create_interval_from_list([row_.chrom,
                                                                           int(row_.start),
                                                                           int(row_.stop)]))
                    if not self.ignore_targets:
                        index.append(i)

                except OverflowError:
                    warnings.warn("""Some of the input sequence were out of range
                                  and have been removed""")
        return intervals, labels[index]

    def __len__(self):
        return len(self.df)

    def _multi_cellular_type(self, list_df):
        multi_df = pd.DataFrame()
        for i, data in enumerate(list_df):
            data['type'] = i + 1
            multi_df = multi_df.append(data)
        return multi_df

    def _binarize_label(self):
        self.ann_df = self.ann_df[self.ann_df.label.isin(self.annotation_list)]
        labels = self.ann_df.label.values

        for i, label in enumerate(self.annotation_list):
            labels[labels == label] = i + 1
        self.ann_df.label = labels

    def _restrict(self):
        assert 'strand' in self.ann_df.columns,\
        'The data need to specify the strand to use restrict'
        df = self.ann_df
        df_plus = pd.DataFrame()
        df_minus = pd.DataFrame()
        if self.predict == 'start':
            df_plus['chrom'] = df[df.strand == '+'].chrom.values
            df_plus['start'] = df[df.strand == '+'].start.values
            df_plus['stop'] = df[df.strand == '+'].start.values + 1
            df_plus['label'] = df[df.strand == '+'].label.values
            df_plus['strand'] = '+'
            df_plus['type'] = df[df.strand == '+'].type.values

            df_minus['chrom'] = df[df.strand == '-'].chrom.values
            df_minus['start'] = df[df.strand == '-'].stop.values - 1
            df_minus['stop'] = df[df.strand == '-'].stop.values
            df_minus['label'] = df[df.strand == '-'].label.values
            df_minus['strand'] = '-'
            df_minus['type'] = df[df.strand == '-'].type.values

        elif self.predict == 'stop':
            df_plus['chrom'] = df[df.strand == '+'].chrom.values
            df_plus['start'] = df[df.strand == '+'].stop.values - 1
            df_plus['stop'] = df[df.strand == '+'].stop.values
            df_plus['label'] = df[df.strand == '+'].label.values
            df_plus['strand'] = '+'
            df_plus['type'] = df[df.strand == '+'].type.values
            
            df_minus['chrom'] = df[df.strand == '-'].chrom.values
            df_minus['start'] = df[df.strand == '-'].start.values
            df_minus['stop'] = df[df.strand == '-'].start.values + 1
            df_minus['label'] = df[df.strand == '-'].label.values
            df_minus['strand'] = '-'
            df_minus['type'] = df[df.strand == '-'].type.values

        self.ann_df = df_plus.append(df_minus)

    def _find_maxlen(self):
        return np.max(self.ann_df.stop.values - self.ann_df.start.values)

    def _calculate_interval(self,
                            df,
                            return_all=False,
                            return_strand=False):
        if return_strand:
            assert 'strand' in df.columns, \
            'To return the strand the dataframe should have a strand column'

        start = df.start.values
        stop = df.stop.values

        if self.data_augmentation and return_all:
            starts = np.concatenate([np.arange(stop[i] - self.length,
                                               start[i] + 1)\
                                     if start[i] - stop[i] + self.length >= 0\
                                     else np.arange(start[i], stop[i] - self.length)\
                                     for i in range(len(start))],
                                     axis=0)
            stops = np.concatenate([np.arange(stop[i],
                                              start[i] + 1 + self.length)\
                                    if start[i] - stop[i] + self.length >= 0\
                                    else np.arange(start[i] + self.length, stop[i])\
                                    for i in range(len(start))],
                                    axis=0)

            if return_strand:
                strand = df.strand.values
                strands = np.concatenate([np.repeat(strand[i],
                                                    start[i] - stop[i] + self.length + 1)\
                                          if stop[i] - start[i] <= self.length
                                          else np.repeat(strand[i],
                                                         stop[i] - start[i] - self.length)
                                          for i in range(len(strand))\
                                           ])
                return starts, stops, strands
            else:
                return starts, stops

        elif self.data_augmentation and not return_all:
            starts = stop - self.length
            stops = start + self.length

            mask = start + self.length - stop <= 0
            starts[mask] = start[mask]
            stops[mask] = stop[mask]
            return starts, stops 

        else:
            wx = (self.length - (stop - start))
            half_wx = wx // 2

            if return_strand:
                return start - half_wx - wx % 2, stop + half_wx, df.strand.values
            else:
                return start - half_wx - wx % 2, stop + half_wx

    def _random_negative_class(self):
        chrom = self.df.chrom.unique()[0]
        number_neg = self.negative_ratio * len(self.df)

        neg_df = pd.DataFrame()
        neg_df['start'] = np.zeros((number_neg,))
        neg_df['stop'] = np.zeros((number_neg,))
        neg_df['chrom'] = chrom
        
        if 'strand' in self.ann_df.columns:
            neg_df['strand'] = np.random.choice(['+', '-'], number_neg)
            
        self.df = self.df.append(neg_df)

        if not self.ignore_targets:
            neg_shape = list(self.labels.shape)
            neg_shape[0] = number_neg
            neg_label = np.zeros(tuple(neg_shape))

            self.labels = np.append(self.labels,
                                    neg_label,
                                    axis=0)

    def _negative_class(self):
        neg_df = pd.DataFrame()
    
        for chrom in self.ann_df.chrom.unique():
            df_ = self.ann_df[self.ann_df.chrom == chrom]
            df_ = df_.sort_values('start')
            pos_starts, pos_stops = self._calculate_interval(df_)
            number_of_pos = np.sum(pos_stops - pos_starts)

            pos_starts = pos_starts - self.length
            list_interval = np.where((pos_starts[1:] - pos_stops[:-1]) > 0)[0]

            if self.negative_ratio == 'all':
                neg_starts = np.concatenate([np.arange(pos_stops[interval_idx],
                                                       pos_starts[interval_idx + 1])\
                                             for interval_idx in list_interval],
                                            0)

            elif isinstance(self.negative_ratio, int):
                length_inters = np.array([pos_starts[i + 1] - pos_stops[i] for i in list_interval])
                proba = length_inters / np.sum(length_inters)

                if self.data_augmentation:
                    number_of_pos *= self.negative_ratio
                else:
                    number_of_pos = len(self.ann_df[self.ann_df.chrom == chrom])
                    number_of_pos *= self.negative_ratio

                interval_chosen = np.random.choice(list_interval,
                                                   number_of_pos,
                                                   p=proba)

                nb_per_interval = [min((interval_chosen == inter).sum(), length_inter)\
                                   for inter, length_inter in zip(list_interval, length_inters)]

                neg_starts = np.concatenate([np.random.choice(np.arange(pos_stops[interval_idx],
                                                                        pos_starts[interval_idx + 1]),
                                                              nb_inter,
                                                              replace=False) for interval_idx,\
                                                                                 nb_inter in\
                                                                                 zip(list_interval,
                                                                                     nb_per_interval)],
                                            0)
            else:
                raise NameError('negative_ratio should be "all" or an integer')

            neg_df_ = pd.DataFrame()
            neg_df_['start'] = neg_starts
            neg_df_['stop'] = neg_starts + self.length
            neg_df_['chrom'] = chrom

            neg_df = neg_df.append(neg_df_)

        neg_df['label'] = 0
        neg_df['type'] = 0

        if 'strand' in self.ann_df.columns:
            neg_df['strand'] = np.random.choice(['+', '-'], len(neg_df))

        nb_types = len(self.ann_df.type.unique())
        nb_labels = len(self.ann_df.label.unique())

        if self.seq2seq:
            labels = np.zeros((len(neg_df),
                                self.length,
                                nb_types,
                                nb_labels))
        else:
            labels = np.zeros((len(neg_df),
                                nb_types,
                                nb_labels))

        return neg_df, labels
    
    def _negative_bed(self):
        neg_df = utils.bed_to_df(self.negative_type, [0])
        incl_chromosomes = self.ann_df.chrom.unique()
        neg_df = neg_df[neg_df.chrom.isin(incl_chromosomes)]

        assert (neg_df.stop.values - neg_df.start.values == self.length).all(),\
        """The negative examples must be of the same length as the positive one"""
        
        neg_df['type'] = 0
        
        if hasattr(self.ann_df, 'strand'):
            assert hasattr(neg_df, 'strand'),\
            """Strand is specifies for positive class and must be specified for
            the negative class as well"""
        else:
            assert not hasattr(neg_df, 'strand'),\
            """Strand is NOT specifies for positive class and must NOT be specified
            for the negative class neither"""
        nb_types = len(self.ann_df.type.unique())
        nb_labels = len(self.ann_df.label.unique())

        if self.seq2seq:
            labels = np.zeros((len(neg_df),
                                self.length,
                                nb_types,
                                nb_labels))
        else:
            labels = np.zeros((len(neg_df),
                                nb_types,
                                nb_labels))
        return neg_df, labels

    def _get_translation_dico(self):
        trans_dico = {}

        for chrom in self.ann_df.chrom.unique():
            df_ = self.ann_df[self.ann_df.chrom == chrom]
            starts = df_.start.values
            stops = df_.stop.values

            if self.data_augmentation:
                len_per_ann = [max(stops[i] - starts[i] - self.length,
                                   starts[i] - stops[i] + self.length + 1) for i in range(len(starts))]
                base_index = np.cumsum(len_per_ann)
                trans_dico[chrom] = {i : np.arange(base_index[i - 1], base_index[i]) \
                                     for i in range(len(starts))}

                trans_dico[chrom][0] = np.arange(base_index[0])
            else:
                trans_dico[chrom] = {i : [i] for i in range(len(starts))}
        return trans_dico

    def _annotation_inter(self):
        inter_dict = {}

        for chrom in self.ann_df.chrom.unique():
            local_df = self.ann_df[self.ann_df.chrom == chrom]
            starts = local_df.start.values
            stops = local_df.stop.values
            mat_ann_start = np.repeat(starts,
                                      len(starts)).reshape((len(starts),
                                                            len(starts)))
            mat_ann_stop = np.repeat(stops,
                                     len(stops)).reshape((len(starts),
                                                          len(starts)))

            if self.defined_positive == 'match_all':
                A = np.minimum(np.abs(mat_ann_stop - mat_ann_stop.T),
                               np.abs(mat_ann_start - mat_ann_start.T))
                m = A.shape[0]
                strided = np.lib.stride_tricks.as_strided
                s0, s1 = A.strides
                out = strided(A.ravel()[1:],
                              shape=(m - 1, m),
                              strides=(s0 + s1, s1)).reshape(m, -1)
            else:
                A = np.minimum(np.abs(mat_ann_stop - mat_ann_stop.T),
                               np.abs(mat_ann_start - mat_ann_start.T),
                               np.abs(mat_ann_start - mat_ann_stop.T))
                m = A.shape[0]
                strided = np.lib.stride_tricks.as_strided
                s0, s1 = A.strides
                out = strided(A.ravel()[1:],
                              shape=(m - 1, m),
                              strides=(s0 + s1, s1)).reshape(m, -1)
            if self.data_augmentation:
                indexes = np.where(out <= self.length)[0]
            else:
                indexes = np.where(out <= self.length // 2)[0]

            inter_indexes = np.unique(indexes)
            inter_dict[chrom] = [np.delete(np.arange(len(starts)), inter_indexes),
                                 inter_indexes]
        return inter_dict

    def _get_dataframe(self):
        new_df = pd.DataFrame()

        for chrom in self.ann_df.chrom.unique():
            df_ = self.ann_df[self.ann_df.chrom == chrom]

            if not self.seq_len == 'real':
                if 'strand' in df_.columns:
                    pos_starts, pos_stops, pos_strands = \
                    self._calculate_interval(df_,
                                             return_all=True,
                                             return_strand=True)

                else:
                    pos_starts, pos_stops = \
                    self._calculate_interval(df_,
                                             return_all=True)

                new_df_ = pd.DataFrame({'start' : pos_starts,
                                        'stop' : pos_stops})
                new_df_['chrom'] = chrom
    
                if 'strand' in df_.columns:
                    new_df_['strand'] = pos_strands

                new_df = new_df.append(new_df_)
            else:
                new_df = new_df.append(df_)
        return new_df

    def _get_labels(self):
        nb_types = len(self.ann_df.type.unique())
        nb_labels = len(self.ann_df.label.unique())
        trans_dict = self._get_translation_dico()
        inter_dict = self._annotation_inter()

        if self.seq2seq:
            labels = np.zeros((1, self.length, nb_types, nb_labels))
        else:
            labels = np.zeros((1, nb_types, nb_labels))

        for chrom in self.ann_df.chrom.unique():
            df_ = self.ann_df[self.ann_df.chrom == chrom]
            pos_starts, pos_stops = self._calculate_interval(df_,
                                                             return_all=True)

            if self.seq2seq:
                labels_ = np.zeros((len(pos_starts),
                                    self.length,
                                    nb_types,
                                    nb_labels))
            else:
                labels_ = np.zeros((len(pos_starts),
                                    nb_types,
                                    nb_labels))
            local_df = self.df[self.df.chrom == chrom]
            ann_to_df = trans_dict[chrom]
            no_inter, inter = inter_dict[chrom]

            if self.seq2seq:
                seq_starts = {idx :
                              np.maximum(np.zeros(len(ann_to_df[idx])).astype(int), 
                                     df_.start.values[idx] - local_df.start.values[ann_to_df[idx]])\
                              for idx in no_inter}
                seq_stops = {idx :
                             np.minimum(np.ones(len(ann_to_df[idx])).astype(int) * self.length,
                                    df_.stop.values[idx] - local_df.start.values[ann_to_df[idx]])\
                             for idx in no_inter}

                labels_[np.concatenate([np.concatenate([np.repeat(ann_to_df[idx][i],
                                                                  seq_stops[idx][i] - seq_starts[idx][i])\
                                                       for i in range(len(ann_to_df[idx]))], 0)\
                                        for idx in no_inter]),
                        np.concatenate([np.concatenate([np.arange(seq_starts[idx][i],
                                                                  seq_stops[idx][i])\
                                                       for i in range(len(ann_to_df[idx]))], 0)\
                                        for idx in no_inter]),
                        np.concatenate([np.repeat(df_.type.values[idx] - 1,
                                                  len(ann_to_df[idx]) *\
                                                  min(self.length,
                                                      df_.stop.values[idx] - df_.start.values[idx]))\
                                        for idx in no_inter]),
                        np.concatenate([np.repeat(df_.label.values[idx] - 1,
                                                  len(ann_to_df[idx]) *\
                                                  min(self.length,
                                                      df_.stop.values[idx] - df_.start.values[idx]))\
                                        for idx in no_inter])] = 1
            else:
                labels_[np.concatenate([ann_to_df[idx] for idx in no_inter], 0),
                        np.concatenate([np.repeat(df_.type.values[idx] - 1,
                                                  len(ann_to_df[idx])) for idx in no_inter], 0),
                        np.concatenate([np.repeat(df_.label.values[idx] - 1,
                                                  len(ann_to_df[idx])) for idx in no_inter], 0)] = 1
            if len(inter) > 0:
                df_inter = np.concatenate([ann_to_df[idx] for idx in inter])
                pos_starts = local_df.start.values[df_inter]
                pos_stops = local_df.stop.values[df_inter]

                ann_starts = df_.start.values[inter]
                ann_stops = df_.stop.values[inter]

                mat_ann_start = np.repeat(ann_starts,
                                          len(pos_starts)).\
                                          reshape(len(ann_starts),
                                          len(pos_starts)).T
                mat_int_start = np.repeat(pos_starts,
                                          len(ann_starts)).\
                                          reshape(len(pos_starts),
                                          len(ann_starts))

                mat_ann_stop = np.repeat(ann_stops,
                                         len(pos_stops)).\
                                         reshape(len(ann_stops),
                                         len(pos_stops)).T
                mat_int_stop = np.repeat(pos_stops,
                                         len(ann_stops)).\
                                         reshape(len(pos_stops),
                                         len(ann_stops))

                if self.seq2seq:
                    idx_int, idx_ann = np.where(np.sign(mat_ann_start - mat_int_stop) *\
                                             np.sign(mat_ann_stop - mat_int_start) < 0)
                    offset_start = mat_ann_start - mat_int_start
                    offset_start[offset_start < 0] = 0 
                    offset_stop = self.length - mat_int_stop + mat_ann_stop
                    offset_stop[offset_stop < 0] = 0

                    for i, j in zip(idx_int, idx_ann):
                        labels_[df_inter[i],
                                offset_start[i, j] : offset_stop[i, j],
                                df_.type.values[inter[j]] - 1,
                                df_.label.values.astype(int)[inter[j]] - 1] = 1

                else:
                    if self.defined_positive == 'match_all':
                        idx_int, idx_ann = np.where(np.sign(mat_ann_start - mat_int_start) *\
                                                    np.sign(mat_ann_stop - mat_int_stop) <= 0)
                    else:
                        idx_int, idx_ann = np.where(np.sign(mat_ann_start - mat_int_stop) *\
                                             np.sign(mat_ann_stop - mat_int_start) < 0)

                    labels_[df_inter[idx_int],
                            df_.type.values[inter[idx_ann]] - 1,
                            df_.label.values[inter[idx_ann]].astype(int) - 1] = 1
            labels = np.append(labels, labels_, axis=0)

        return labels[1:]

    @property
    def command_dict(self):
        return utils.ArgumentsDict(self, kwargs=False)


class ContinuousDataset(object):
    """
    info:
        docs: >
            Reads files adaptated for continuous annotation (wig, BigWig,
            bedGraph), and returns intervals and the corresponding annotation
            as a label.
            
            An interval can be labeled with two manners. First, the label is
            the experimental values on a window at the center of the interval
            (window of any length within the interval). Secondly, it can be
            labeled by the experimental values covering all the interval and
            downsampled to reach a smaller length. Downsampling can be achived
            by taking one value from several ones or by averaging the values
            within small window.
            Note that wig and bedGraph will be converted and a file with the
            chromosomes size is needed.
    
    args:
        annotation_files:
            list of file with annotations (wig, bigWig or bedGraph).
            If we just want the inputs then a file finishing by .sizes can be
            passed (it must contains the size of chromosome).
         window:
            the length of the intervals.
        tg_window:
            the length of the target window (should be a divisor of the window
            length if downsampling). default=1
        nb_annotation_type:
            The number of different annotation in input files. The same number
            of files must be passed for every annotation.
            The list must be organised as [file1_ann1, file1_ann2, file2_ann1,
            file2_ann2], with file1, file2 designing two differents kind of
            files (different lab, different cellular type ...).
            If None the output shape will be (batch_size, tg_window, nb_of_file)
        downsampling:
            {None, 'mean', 'downsampling'}, how the label is created, if None
            the label is the original values at the center of the interval, if
            'mean' downsampling by averaging on N values recursively,
            if 'downsampling' taking the first value every N values.
            default=None
        normalization_mode:
            arguments from Normalizer class to normalize the data.
            default=None
        overlapping:
            boolean or int, weither or not to return all the possible intervals, if
            False only the intervals corresponding to non overlapping target will
            be returned, if int the stride between the interval will be of int.
            default=True.
        num_chr:
            if specified, 'chr' in the chromosome name will be dropped,
            default=False
        incl_chromosomes:
            exclusive list of chromosome names to include in the final dataset.
            if not None, only these will be present in the dataset,
            default=None
        excl_chromosomes:
            list of chromosome names to omit from the dataset. default=None
        start_stop:
            list of tuple indicating where to start and stop generating data.
            One per included chromosome. default=None
        ignore_targets: 
            if True, target variables are ignored, default=False
        size:
            A file with the chromosome name and size usefull to convert wig
            and bedGraph to bigwig.
            default= None
        path_to_ucsc:
            The path to the ucsc utils (wigToBigWig and bedGraphToBigWig) to
            give if those functionality can not be downloaded with conda.
            default=None
    """
    def __init__(self, annotation_files,
                       window,
                       tg_window=1,
                       nb_annotation_type=None,
                       downsampling=None,
                       normalization_mode=None,
                       overlapping=True,
                       num_chr=False,
                       incl_chromosomes=None,
                       excl_chromosomes=None,
                       start_stop=None,
                       ignore_targets=False,
                       size=None,
                       path_to_ucsc=None):
        
        self.annotation_files = annotation_files
        self.nb_annotation_type = nb_annotation_type
        self.window = window
        self.hw = window // 2
        self.tg_window = tg_window
        self.downsampling = downsampling
        self.normalization_mode = normalization_mode
        self.overlapping = overlapping
        self.num_chr = num_chr
        self.incl_chromosomes = incl_chromosomes
        self.excl_chromosomes = excl_chromosomes
        self.start_stop = start_stop
        self.ignore_targets = ignore_targets
        self.df = pd.DataFrame()
        self.size = size
        self.path_to_ucsc = path_to_ucsc
        self.frame = inspect.currentframe()

        # converting to list type to consistancy with the case of multi-outputs
        if not isinstance(self.annotation_files, list):
            self.annotation_files = [self.annotation_files]
        self.chrom_size = dict()
        
        for idx, annotation_file in enumerate(self.annotation_files):
            if annotation_file.endswith('.wig'):
                assert self.size is not None,\
                '''To use wig file a file with the chromosome size must be
                parsed a size'''
                utils.wig_to_df(annotation_file,
                                self.size,
                                self.path_to_ucsc)
                self.annotation_files[idx] = annotation_file[:-3] + 'bw'
            if annotation_file.endswith('.bedGraph'):
                assert self.size is not None,\
                '''To use bedGraph file a file with the chromosome size must be
                parsed a size'''
                utils.bedGraph_to_df(annotation_file,
                                     self.size,
                                     self.path_to_ucsc)
                self.annotation_files[idx] = annotation_file[:-8] + 'bw'

        if self.annotation_files[0].endswith(('.wig', '.bw', 'bedGraph')):
            bw = pyBigWig.open(self.annotation_files[0])
            # omit data outside chromosomes
            if incl_chromosomes is not None:
                for name, size in bw.chroms().items():
                    if name in incl_chromosomes:
                        self.chrom_size[name] = size
                        
            elif excl_chromosomes is not None:
                for name, size in bw.chroms().items():
                    if name not in excl_chromosomes:
                        self.chrom_size[name] = size
            else:
                self.chrom_size = bw.chroms()
            bw.close()

        elif self.annotation_files[0].endswith('.sizes'):
            chrom_size = pd.read_csv(self.annotation_files[0],
                                     sep='\t',
                                     names=['chrom', 'sizes'])

            if incl_chromosomes is not None:
                for name, size in zip(chrom_size.chrom.values,
                                      chrom_size.sizes.values):
                    if name in incl_chromosomes:
                        self.chrom_size[name] = size
    
            elif excl_chromosomes is not None:
                for name, size in zip(chrom_size.chrom.values,
                                      chrom_size.sizes.values):
                    if name not in excl_chromosomes:
                        self.chrom_size[name] = size
        
        if self.start_stop is not None:
            if isinstance(self.start_stop, tuple):
                self.start_stop = [self.start_stop]
            assert len(self.start_stop) == len(self.chrom_size),\
            """len(start_stop) should match the number of chromosome
            included. Found {} and expected {}""".format(len(self.start_stop),
                                                         len(self.chrom_size))

            for size, start_stop in zip(self.chrom_size.values(), self.start_stop):
                assert start_stop[0] >= self.hw,\
                """start in start_stop must be greater than the half of window"""
                assert start_stop[1] < size - self.hw,\
                """stop in start_stop must be smaller than the chromosome size
                minus the half of window"""

        self.asteps = 1
        if isinstance(self.overlapping, int):
            self.asteps = self.overlapping

        if not self.downsampling:
            if not self.overlapping:
                self.asteps = self.tg_window
        else:
            if not self.overlapping:
                self.asteps = self.window

        self.df = self._get_dataframe()

        if not self.ignore_targets:
            self.extractor = bbi_extractor(self.annotation_files,
                                           self.tg_window,
                                           self.nb_annotation_type,
                                           self.downsampling,
                                           self.normalization_mode)

        if self.num_chr and self.df.iloc[0][0].startswith("chr"):
            self.df.chrom = self.df.chrom.str.replace("^chr", "")
        if not self.num_chr and not self.df.iloc[0][0].startswith("chr"):
            self.df.chrom = "chr" + self.df.chrom

    @classmethod
    def default_dict(cls):
        return utils.get_default_args(cls.__init__)

    @classmethod
    def predict_label_shape(cls, **input_dict):
        command_dict = cls.default_dict()
        command_dict.update(input_dict)

        if command_dict['ignore_targets']:
            return None

        assert 'annotation_files' in command_dict,\
        """To create an instance passing annotation_files is required"""

        assert 'window' in command_dict,\
        """To create an instance passing window is required"""

        if isinstance(command_dict['annotation_files'], str):
            command_dict['annotation_files'] = [command_dict['annotation_files']]
        
        if isinstance(command_dict['nb_annotation_type'], int):
            assert len(command_dict['annotation_files'])\
            % command_dict['nb_annotation_type'] == 0,\
            """nb_annotation_type should devide the number of annotation files"""
            return (command_dict['tg_window'],
                    len(command_dict['annotation_files']) // command_dict['nb_annotation_type'],
                    command_dict['nb_annotation_type'])
        else:
            return (command_dict['tg_window'],
                    len(command_dict['annotation_files']))

    @property
    def label_shape(self):
        command_dict = self.command_dict.as_input()
        return self.predict_label_shape(**command_dict)

    def _get_dataframe(self):
        chrom = list()
        start = list()
        stop = list()
        first_index = list()
        last_index = list()

        if self.start_stop is not None:
            for name, start_stop in zip(self.incl_chromosomes,
                                        self.start_stop):
                chrom.append(name)
                start.append(start_stop[0])
                stop.append(start_stop[1])
                first_index.append(0)
                last_index.append((stop[-1] - start[-1]) // self.asteps)
        else:
            for name, size in self.chrom_size.items():
                chrom.append(name)
                start.append(self.hw)
                stop.append(size - self.hw - (self.window % 2))
                first_index.append(0)
                last_index.append((stop[-1] - start[-1]) // self.asteps)

        last_index = np.cumsum(last_index)
        for i in range(len(last_index)):
            last_index[i] += i
        first_index[1:] = last_index[:-1] + 1

        new_df = pd.DataFrame({'chrom' : chrom,
                               'start' : start,
                               'stop' : stop,
                               'first_index' : first_index,
                               'last_index' : last_index})
        return new_df

    def _get_interval(self, idx):
        indicative_mat = (np.sign(self.df.first_index.values - idx)) *\
                         (np.sign(self.df.last_index.values - idx))
        df_idx = np.where(indicative_mat <= 0)[0][-1]

        row = self.df.iloc[df_idx]
        start = row.start + (idx - row.first_index) * self.asteps - self.hw
        stop = row.start + (idx - row.first_index) * self.asteps + self.hw + (self.window % 2)
        interval = pybedtools.create_interval_from_list([row.chrom,
                                                         int(start),
                                                         int(stop)])
        return interval

    def __getitem__(self, idx):
        """Returns (pybedtools.Interval, labels)"""
        if not isinstance(idx, list):
            idx = [idx]

        intervals = [self._get_interval(index) for index in idx]

        if self.ignore_targets:
            labels = {}
        else:
            labels = np.array([self.extractor.extract(interval) for interval in intervals])

        return intervals, labels

    def __len__(self):
        return self.df.last_index.values[-1] + 1

    @property
    def command_dict(self):
        return utils.ArgumentsDict(self, kwargs=False)


class StringSeqIntervalDl(object):
    """
    info:
        docs: >
            Dataloader for a combination of fasta and a file with annotations.
            The dataloader extracts regions from the fasta file corresponding
            to the `annotation_file`. Returned sequences are of the type
            np.array([str]), possibly the corresponding occupancy taken from a
            bbi file can be passed as secondary input or targets.
    args:
        annotation_files:
            list of file with annotations (wig, bigWig or bedGraph / bed, gff)
        fasta_file:
            Reference genome FASTA file path.
        force_upper:
            Force uppercase output of sequences
        use_strand:
            boolean, whether or not to respect the strand for spare annotation.
            If false all the sequence are ridden in 5'.
        sec_inputs:
            Path to other bbi files, the corresponding coverage on the interval
            will be used as a secondary input for the model (or targets)
            default=None
        sec_input_length:
            {int, 'maxlen'}, Length of the secondary sequences to be used as
            model input. If maxlen the length will be the same as the DNA seq.
            default='maxlen'
        sec_input_shape:
            The shape of the secondary inputs, in order to adapt this shape to
            the need.
            default=None.
        sec_nb_annotation:
            The number of different annotation in secondary input files.
            (see ContinuousDataset for details).
            default=None
        sec_sampling_mode:
            How the secondary inputs are sampled from the coverage on the cor-
            -responding input interval.
            default=None
        sec_normalization_mode:
            How the secondary inputs are normalized.
        use_sec_as:
            {'inputs', 'targets'}
            default='inputs'
        rc:
            boolean, if true the batch is reversed complemented.
            default=False
        args: 
            Arguments to be passed to the dataset reader
        kwargs: 
           Dictionnary of arguments specific to the dataset reader
    output_schema:
        inputs:
            name: seq
            shape: ()
            doc: DNA sequence as string
            special_type: DNAStringSeq
            associated_metadata: ranges
        targets:
            shape: (None,)
            doc: (optional) values corresponding to the annotation file
    """
    def __init__(self,
                 annotation_files,
                 fasta_file,
                 use_strand=False,
                 sec_inputs=None,
                 sec_input_length='maxlen',
                 sec_input_shape=None,
                 sec_nb_annotation=None,
                 sec_sampling_mode=None,
                 sec_normalization_mode=None,
                 use_sec_as='inputs',
                 force_upper=False,
                 rc=False,
                 *args,
                 **kwargs):
        self.annotation_files = annotation_files
        self.fasta_file = fasta_file
        self.use_strand = use_strand
        self.force_upper = force_upper
        self.fasta_extractors = None
        self.pad_seq = False
        self.sec_inputs = sec_inputs
        self.sec_input_length = sec_input_length
        self.sec_input_shape = sec_input_shape
        self.sec_nb_annotation = sec_nb_annotation
        self.sec_sampling_mode = sec_sampling_mode
        self.sec_normalization_mode = sec_normalization_mode
        self.use_sec_as = use_sec_as
        self.rc = rc
        self.frame = inspect.currentframe()

        assert self.use_sec_as in ['targets', 'inputs'],\
        'use_sec_as is either "targets" or "input"'

        if not isinstance(self.annotation_files, list):
            self.annotation_files = [self.annotation_files]

        if self.annotation_files[0].endswith(('.bed', '.gff', 'gff3', 'gtf')):
            self.dataset = SparseDataset(annotation_files = self.annotation_files,
                                         *args,
                                         **kwargs)

            if self.dataset.seq_len == 'real':
                self.pad_seq = True

        elif self.annotation_files[0].endswith(('.wig', '.bw', 'bedGraph', '.sizes')):
            self.dataset = ContinuousDataset(annotation_files = self.annotation_files,
                                             *args,
                                             **kwargs)
        if self.sec_input_length == 'maxlen':
            try:
                self.sec_input_length = self.dataset.length
            except AttributeError:
                self.sec_input_length = self.dataset.window

        if self.sec_inputs:
            self.extractor = bbi_extractor(self.sec_inputs,
                                           self.sec_input_length,
                                           self.sec_nb_annotation,
                                           self.sec_sampling_mode,
                                           self.sec_normalization_mode)

    @classmethod
    def default_dict(cls):
        return utils.get_default_args(cls.__init__)

    @classmethod
    def predict_sec_input_shape(cls, **input_dict):
        command_dict = cls.default_dict()
        command_dict.update(input_dict)

        if command_dict['sec_inputs'] is None:
            return None

        assert not isinstance(command_dict['sec_input_length'], str),\
        """To anticipate the secondary shape sec_input_length must be an integer"""

        if isinstance(command_dict['sec_inputs'], str):
            command_dict['sec_inputs'] = [command_dict['sec_inputs']]

        if isinstance(command_dict['sec_nb_annotation'], int):
            assert len(command_dict['sec_inputs'])\
            % command_dict['sec_nb_annotation'] == 0,\
            """sec_nb_annotation should devide the length of sec_inputs"""
            
            sec_input_shape = (command_dict['sec_input_length'],
                               len(command_dict['sec_inputs'])\
                               // command_dict['sec_nb_annotation'],
                               command_dict['sec_nb_annotation'])
        else:
            sec_input_shape =  (command_dict['sec_input_length'],
                                len(command_dict['sec_inputs']))
            
        if command_dict['sec_input_shape']:
            try:
                np.zeros(sec_input_shape).reshape(command_dict['sec_input_shape'][1:])
                return command_dict['sec_input_shape'][1:]
            except ValueError:
                raise ValueError("""The required secondary input shape is not compatible with the 
                other arguments""")
                
        else:
            return sec_input_shape

    @property
    def secondary_input_shape(self):
        command_dict = self.command_dict.as_input()
        command_dict['sec_input_length'] = self.sec_input_length
        return self.predict_sec_input_shape(**command_dict)

    @classmethod
    def predict_label_shape(cls, **input_dict):
        assert 'annotation_files' in input_dict,\
        """To create an instance passing annotation_files is required"""

        if isinstance(input_dict['annotation_files'], str):
            input_dict['annotation_files'] = [input_dict['annotation_files']]
        
        if input_dict['annotation_files'][0].endswith(('.bed', '.gff', 'gff3', 'gtf')):
            return SparseDataset.predict_label_shape(**input_dict)
        elif input_dict['annotation_files'][0].endswith(('.wig', '.bw', 'bedGraph')):
            return ContinuousDataset.predict_label_shape(**input_dict)

    @property
    def label_shape(self):
        return self.dataset.label_shape

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        if not isinstance(idx, list):
            idx = [idx]

        self.fasta_extractors = FastaStringExtractor(self.fasta_file,
                                                     use_strand=self.use_strand,
                                                     force_upper=self.force_upper)

        intervals, labels = self.dataset[idx]
        seqs = list()

        if self.use_strand:
            negative_strand = list()

            assert hasattr(intervals[0], 'strand'),\
            '''Strand need to be specified to use use_strand'''

            for i in range(len(intervals)):
                interval = intervals[i]
                if interval.length == 0:
                    seqs.append(''.join(random.choices('ATGC',
                                                       k=self.dataset.length)))
                elif interval.strand == '-':
                    seqs.append(self.fasta_extractors.extract(interval))
                    negative_strand.append(i)
                else:
                    seqs.append(self.fasta_extractors.extract(interval))

            if self.dataset.seq2seq == True:
                labels[negative_strand] = labels[negative_strand, ::-1, :, :]

        else:  
            for interval in intervals:
                if interval.length == 0:
                    seqs.append(''.join(random.choices('ATGC',
                                                       k=self.dataset.length)))
                else:
                    seqs.append(self.fasta_extractors.extract(interval) )

        if self.pad_seq:
                seqs = [fixed_len(seq,
                             int(self.dataset.length),
                             anchor="center",
                             value="N") for seq in seqs]

        if self.sec_inputs:
            sec_seqs = [self.extractor.extract(interval) for interval in intervals]
            sec_seqs = np.array(sec_seqs)
            
            if self.use_strand:
                sec_seqs[negative_strand] = sec_seqs[negative_strand, ::-1]            

            if self.rc:
                seqs, sec_seqs, labels = utils.reverse_complement(seqs,
                                                                 labels,
                                                                 sec_seqs)
            if self.sec_input_shape:
                sec_seqs = sec_seqs.reshape((sec_seqs.shape[0],) + \
                                             self.sec_input_shape[1:])
            if self.use_sec_as == 'inputs':
                inputs = [np.array(seqs), sec_seqs]
            else:
                inputs = np.array(seqs)
                labels = [labels, sec_seqs]
        else:
            if self.rc:
                seqs, labels = utils.reverse_complement(seqs, labels)
            inputs = np.array(seqs)

        return {
            "inputs": inputs,
            "targets": labels,
            }

    @property
    def command_dict(self):
        return utils.ArgumentsDict(self, called_args='dataset')


class SeqIntervalDl(object):
    """
    info:
        doc: >
            Dataloader for a combination of fasta and tab-delimited input files
            such as bed files. The dataloader extracts regions from the fasta
            file corresponding to the `annotation_file` and converts them into
            one-hot encoded format. Returned sequences are of the type np.array
            with the shape inferred from the arguments: `alphabet_axis` and
            `dummy_axis`.
    args:
        alphabet_axis:
            doc: axis along which the alphabet runs (e.g. A,C,G,T for DNA)
        dummy_axis:
            doc: defines in which dimension a dummy axis should be added.
            None if no dummy axis is required.
        alphabet:
            doc: >
                alphabet to use for the one-hot encoding. This defines the
                order of the one-hot encoding.
                Can either be a list or a string: 'ACGT' or ['A, 'C', 'G', 'T'].
                Default: 'ACGT'
        dtype:
            doc: 'defines the numpy dtype of the returned array.
            Example: int, np.int32, np.float32, float'
        args: 
            arguments specific to the different dataloader that can be used
        kwargs: 
            dictionnary with specific arguments to the dataloader.
    output_schema:
        inputs:
            name: seq
            shape: (None, 4)
            doc: One-hot encoded DNA sequence
            special_type: DNASeq
            associated_metadata: ranges
        targets:
            shape: (None,)
            doc: (optional) values given in the annotation file
    """
    def __init__(self,
                 alphabet_axis=1,
                 dummy_axis=None,
                 alphabet=DNA,
                 dtype=None,
                 *args,
                 **kwargs):
        self.frame = inspect.currentframe()
        # core dataset, not using the one-hot encoding params
        self.seq_dl = StringSeqIntervalDl(*args,
                                          **kwargs)

        self.input_transform = ReorderedOneHot(alphabet=alphabet,
                                               dtype=dtype,
                                               alphabet_axis=alphabet_axis,
                                               dummy_axis=dummy_axis)

    def __len__(self):
        return len(self.seq_dl)

    def __getitem__(self, idx):
        ret = self.seq_dl[idx]
        
        if self.seq_dl.sec_inputs and self.seq_dl.use_sec_as == 'inputs':
            length = len(ret['inputs'][0])
            ret['inputs'] = [np.array([self.input_transform(str(ret["inputs"][0][i]))\
                             for i in range(length)]), ret['inputs'][1]]
        else:   
            length = len(ret['inputs'])
            ret['inputs'] = np.array([self.input_transform(str(ret["inputs"][i]))\
                            for i in range(length)])
        return ret

    @property
    def command_dict(self):
        return utils.ArgumentsDict(self, called_args='seq_dl')

    @classmethod
    def default_dict(cls):
        return utils.get_default_args(cls.__init__)
    
    @classmethod
    def predict_sec_input_shape(cls, **input_dict):
        return StringSeqIntervalDl.predict_sec_input_shape(**input_dict)

    @property
    def secondary_input_shape(self):
        return self.seq_dl.secondary_input_shape
    
    @classmethod
    def predict_label_shape(cls, **input_dict):
        return StringSeqIntervalDl.predict_label_shape(**input_dict)
    
    @property
    def label_shape(self):
        command_dict = self.command_dict.as_input()
        return self.seq_dl.label_shape
    
    @classmethod
    def predict_input_shape(cls, **input_dict):
        command_dict = cls.default_dict()
        command_dict.update(input_dict)
        
        assert "annotation_files" in command_dict,\
        """annotation_files is needed to calculate the input shape"""

        if isinstance(command_dict['annotation_files'], str):
            command_dict['annotation_files'] = [command_dict['annotation_files']]

        if command_dict['annotation_files'][0].endswith(('.bed', '.gff', 'gff3', 'gtf')):
            assert 'seq_len' in command_dict,\
            """seq_len can not be set as default if we want to anticipate the input shape"""
            assert not isinstance(command_dict['seq_len'], str),\
            """seq_len must be an integer to calculate the input shape"""
            length = command_dict['seq_len']
            
        elif command_dict['annotation_files'][0].endswith(('.wig', '.bw', 'bedGraph')):
            assert 'window' in command_dict,\
            """window is needed to calculate the input shape with bigwig files"""
            length = command_dict['window']
            
        if command_dict['dummy_axis']:
            shape = np.zeros((3,), dtype=int)
            shape[command_dict['dummy_axis']] = 1
            shape[command_dict['alphabet_axis']] = 4
            shape[shape == 0] = length
            return tuple(shape)
        else:
            shape = np.zeros((2,), dtype=int)
            shape[command_dict['alphabet_axis']] = 4
            shape[shape == 0] = length
            return tuple(shape)
    
    @property
    def input_shape(self):
        command_dict = self.command_dict.as_input()

        if command_dict['annotation_files'][0].endswith(('.bed', '.gff', 'gff3', 'gtf')):
            command_dict['seq_len'] = self.seq_dl.dataset.length

        return self.predict_input_shape(**command_dict)
    


from __future__ import absolute_import

from . import evaluation
from . import normalization
from . import extractors
from . import generators
from . import layers
from . import model
from . import utils
from . import sequence


# Also importable from root
from .generators import Generator, MultiGenerator
from .model import ModelWrapper
from .sequence import SeqIntervalDl, StringSeqIntervalDl

__version__ = '0.0.46'


def get_ucsc():
    import os, stat
    distrib = ''
    while distrib not in {"linux.x86_64.v369", 
                           "linux.x86_64.v385",
                           "linux.x86_64",
                           "macOSX.x86_64"}:
        distrib = input("Choose your OS: (linux.x86_64.v369," + \
        "linux.x86_64.v385, linux.x86_64, macOSX.x86_64)\n")
    
    path = input("Enter the path to the directory where you want to save the UCSC utils.\n")

    if not os.path.exists(path):
        os.mkdir(path)

    os.system("wget -O " + path + '/wigToBigWig https://hgdownload.soe.ucsc.edu/admin/exe/' + distrib + '/wigToBigWig')
    os.system("wget -O " + path + '/bedGraphToBigWig https://hgdownload.soe.ucsc.edu/admin/exe/' + distrib + '/bedGraphToBigWig')

    os.chmod(path + '/wigToBigWig', stat.S_IEXEC)
    os.chmod(path + '/bedGraphToBigWig', stat.S_IEXEC)



import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent
README = (HERE/"README.md").read_text()

setup(name="keras_dna",
      version="0.0.46",
      description="Build keras generator for genomic application",
      long_description=README,
      long_description_content_type="text/markdown",
      url="https://github.com/etirouthier/keras_dna",
      author="Etienne Routhier",
      author_email="etienne.routhier@upmc.fr",
      license="MIT",
      classifiers=[
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6"],
      packages=find_packages(),
      include_packages_data=True,
      install_requires=["tensorflow>=2.0.0",
                        "numpy",
                        "pandas",
                        "kipoiseq",
                        "pyBigWig",
                        "pybedtools"]
      )
      
